from typing import (
    Tuple,
)

from eth_typing import (
    Address,
)

from eth import (
    constants,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    OutOfBoundsRead,
)


def balance(computation: ComputationAPI) -> None:
    addr = force_bytes_to_address(computation.stack_pop1_bytes())
    push_balance_of_address(addr, computation)


def selfbalance(computation: ComputationAPI) -> None:
    push_balance_of_address(computation.msg.storage_address, computation)


def push_balance_of_address(address: Address, computation: ComputationAPI) -> None:
    balance = computation.state.get_balance(address)
    computation.stack_push_int(balance)


def origin(computation: ComputationAPI) -> None:
    computation.stack_push_bytes(computation.transaction_context.origin)


def address(computation: ComputationAPI) -> None:
    computation.stack_push_bytes(computation.msg.storage_address)


def caller(computation: ComputationAPI) -> None:
    computation.stack_push_bytes(computation.msg.sender)


def callvalue(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.msg.value)


def calldataload(computation: ComputationAPI) -> None:
    """
    Load call data into memory.
    """
    start_position = computation.stack_pop1_int()

    value = computation.msg.data_as_bytes[start_position : start_position + 32]
    padded_value = value.ljust(32, b"\x00")
    normalized_value = padded_value.lstrip(b"\x00")

    computation.stack_push_bytes(normalized_value)


def calldatasize(computation: ComputationAPI) -> None:
    size = len(computation.msg.data)
    computation.stack_push_int(size)


def calldatacopy(computation: ComputationAPI) -> None:
    (
        mem_start_position,
        calldata_start_position,
        size,
    ) = computation.stack_pop_ints(3)

    computation.extend_memory(mem_start_position, size)

    word_count = ceil32(size) // 32
    copy_gas_cost = word_count * constants.GAS_COPY

    computation.consume_gas(copy_gas_cost, reason="CALLDATACOPY fee")

    value = computation.msg.data_as_bytes[
        calldata_start_position : calldata_start_position + size
    ]
    padded_value = value.ljust(size, b"\x00")

    computation.memory_write(mem_start_position, size, padded_value)


def chain_id(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.state.execution_context.chain_id)


def codesize(computation: ComputationAPI) -> None:
    size = len(computation.code)
    computation.stack_push_int(size)


def codecopy(computation: ComputationAPI) -> None:
    (
        mem_start_position,
        code_start_position,
        size,
    ) = computation.stack_pop_ints(3)

    computation.extend_memory(mem_start_position, size)

    word_count = ceil32(size) // 32
    copy_gas_cost = constants.GAS_COPY * word_count

    computation.consume_gas(
        copy_gas_cost,
        reason="CODECOPY: word gas cost",
    )

    with computation.code.seek(code_start_position):
        code_bytes = computation.code.read(size)

    padded_code_bytes = code_bytes.ljust(size, b"\x00")

    computation.memory_write(mem_start_position, size, padded_code_bytes)


def gasprice(computation: ComputationAPI) -> None:
    computation.stack_push_int(computation.transaction_context.gas_price)


def extcodesize(computation: ComputationAPI) -> None:
    account = force_bytes_to_address(computation.stack_pop1_bytes())
    code_size = len(computation.state.get_code(account))

    computation.stack_push_int(code_size)


def extcodecopy_execute(computation: ComputationAPI) -> Tuple[Address, int]:
    """
    Runs the logical component of extcodecopy, without charging gas.

    :return (target_address, copy_size): useful for the caller to determine gas costs
    """
    account = force_bytes_to_address(computation.stack_pop1_bytes())
    (
        mem_start_position,
        code_start_position,
        size,
    ) = computation.stack_pop_ints(3)

    computation.extend_memory(mem_start_position, size)

    code = computation.state.get_code(account)

    code_bytes = code[code_start_position : code_start_position + size]
    padded_code_bytes = code_bytes.ljust(size, b"\x00")

    computation.memory_write(mem_start_position, size, padded_code_bytes)

    return account, size


def consume_extcodecopy_word_cost(computation: ComputationAPI, size: int) -> None:
    word_count = ceil32(size) // 32
    copy_gas_cost = constants.GAS_COPY * word_count
    computation.consume_gas(
        copy_gas_cost,
        reason="EXTCODECOPY: word gas cost",
    )


def extcodecopy(computation: ComputationAPI) -> None:
    _address, size = extcodecopy_execute(computation)
    consume_extcodecopy_word_cost(computation, size)


def extcodehash(computation: ComputationAPI) -> None:
    """
    Return the code hash for a given address.
    EIP: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1052.md
    """
    account = force_bytes_to_address(computation.stack_pop1_bytes())
    state = computation.state

    if state.account_is_empty(account):
        computation.stack_push_bytes(constants.NULL_BYTE)
    else:
        computation.stack_push_bytes(state.get_code_hash(account))


def returndatasize(computation: ComputationAPI) -> None:
    size = len(computation.return_data)
    computation.stack_push_int(size)


def returndatacopy(computation: ComputationAPI) -> None:
    (
        mem_start_position,
        returndata_start_position,
        size,
    ) = computation.stack_pop_ints(3)

    if returndata_start_position + size > len(computation.return_data):
        raise OutOfBoundsRead(
            "Return data length is not sufficient to satisfy request.  Asked "
            f"for data from index {returndata_start_position} "
            f"to {returndata_start_position + size}.  "
            f"Return data is {len(computation.return_data)} bytes in length."
        )

    computation.extend_memory(mem_start_position, size)

    word_count = ceil32(size) // 32
    copy_gas_cost = word_count * constants.GAS_COPY

    computation.consume_gas(copy_gas_cost, reason="RETURNDATACOPY fee")

    value = computation.return_data[
        returndata_start_position : returndata_start_position + size
    ]

    computation.memory_write(mem_start_position, size, value)



from rlp.sedes import (
    CountableList,
)

from eth.rlp.headers import (
    BlockHeader,
)
from eth.vm.forks.homestead.blocks import (
    HomesteadBlock,
)

from .transactions import (
    SpuriousDragonTransaction,
)


class SpuriousDragonBlock(HomesteadBlock):
    transaction_builder = SpuriousDragonTransaction
    fields = [
        ("header", BlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(BlockHeader)),
    ]



from .clique import (  # noqa: F401
    CliqueApplier,
    CliqueConsensus,
    CliqueConsensusContext,
)
from .constants import (  # noqa: F401
    NONCE_AUTH,
    NONCE_DROP,
)
from .datatypes import (  # noqa: F401
    Snapshot,
    Tally,
    Vote,
    VoteAction,
)



from .vm import (  # noqa: F401
    fill_vm_test,
)
from .state import (  # noqa: F401
    fill_state_test,
)
from .common import (  # noqa: F401
    execution,
    expect,
    pre_state,
    setup_filler,
    setup_main_filler,
)
from .main import (  # noqa: F401
    fill_test,
)



from abc import (
    ABC,
)
from typing import (
    Type,
)

from eth_utils import (
    encode_hex,
)
from rlp.sedes import (
    CountableList,
)

from eth.abc import (
    TransactionBuilderAPI,
)

from ..london import (
    LondonBlock,
)
from ..london.blocks import (
    LondonBackwardsHeader,
    LondonBlockHeader,
    LondonMiningHeader,
)
from .transactions import (
    ArrowGlacierTransactionBuilder,
)


class ArrowGlacierMiningHeader(LondonMiningHeader, ABC):
    pass


class ArrowGlacierBlockHeader(LondonBlockHeader, ABC):
    def __str__(self) -> str:
        return f"<ArrowGlacierBlockHeader #{self.block_number} {encode_hex(self.hash)[2:10]}>"  # noqa: E501


class ArrowGlacierBlock(LondonBlock):
    transaction_builder: Type[TransactionBuilderAPI] = ArrowGlacierTransactionBuilder
    fields = [
        ("header", ArrowGlacierBlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(LondonBackwardsHeader)),
    ]



from typing import (
    FrozenSet,
    List,
    NamedTuple,
    Set,
)

from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    encode_hex,
    get_extended_debug_logger,
    int_to_big_endian,
    to_bytes,
    to_int,
)
import rlp
from trie import (
    HexaryTrie,
    exceptions as trie_exceptions,
)
from web3 import Web3
from web3.types import (
    BlockIdentifier,
)

from eth._utils.padding import (
    pad32,
)
from eth.abc import (
    AccountStorageDatabaseAPI,
    AtomicDatabaseAPI,
    DatabaseAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
)
from eth.db.backends.base import (
    BaseDB,
)
from eth.db.backends.memory import (
    MemoryDB,
)
from eth.db.batch import (
    BatchDB,
)
from eth.db.cache import (
    CacheDB,
)
from eth.db.journal import (
    JournalDB,
)
from eth.typing import (
    JournalDBCheckpoint,
)
from eth.vm.interrupt import (
    MissingStorageTrieNode,
)

ZERO_VALUE = b'\x00' * 32

class PendingWrites(NamedTuple):
    """
    A set of variables captured just before account storage deletion.
    The variables are used to revive storage if the EVM reverts to a point
    prior to deletion.
    """

    write_trie: HexaryTrie  # The write trie at the time of deletion
    trie_nodes_batch: BatchDB  # A batch of all trie nodes written to the trie
    starting_root_hash: Hash32  # The starting root hash


class StorageLookup(BaseDB):
    """
    This lookup converts lookups of storage slot integers into the appropriate trie
    lookup. Similarly, it persists changes to the appropriate trie at write time.

    StorageLookup also tracks the state roots changed since the last persist.
    """

    logger = get_extended_debug_logger("eth.db.storage.StorageLookup")

    # The trie that is modified in-place, used to calculate storage root on-demand
    _write_trie: HexaryTrie

    # These are the new trie nodes, waiting to be committed to disk
    _trie_nodes_batch: BatchDB

    # When deleting an account, push the pending write info onto this stack.
    # This stack can get as big as the number of transactions per block: one for
    # each delete.
    _historical_write_tries: List[PendingWrites]

    _w3: Web3
    _block_identifier: BlockIdentifier
    _address: Address

    def __init__(
            self,
            db: DatabaseAPI,
            w3: Web3,
            block_identifier: BlockIdentifier,
            address: Address
        ) -> None:
        self._db = db
        self._w3 = w3
        self._block_identifier = block_identifier
        self._address = address

    def _make_db_key(self, key: bytes) -> bytes:
        return b'storage' + bytes(self._address) + key

    def __getitem__(self, key: bytes) -> bytes:
        k = self._make_db_key(key)

        if k not in self._db:
            val = self._w3.eth.get_storage_at(
                self._address,
                to_int(key),
                block_identifier=self._block_identifier
            )
            self._db[k] = val
        
        return self._db[k]

    def __setitem__(self, key: bytes, value: bytes) -> None:
        raise NotImplementedError("StorageLookup does not support __setitem__")
        k = self._make_db_key(key)
        self._db[k] = value

    def _exists(self, key: bytes) -> bool:
        # used by BaseDB for __contains__ checks
        raise NotImplementedError("StorageLookup does not support __contains__")
        hashed_slot = self._decode_key(key)
        read_trie = self._get_read_trie()
        return hashed_slot in read_trie

    def __delitem__(self, key: bytes) -> None:
        k = self._make_db_key(key)
        self._db[k] = ZERO_VALUE

    def commit_to(self, db: DatabaseAPI) -> None:
        """
        Trying to commit changes when nothing has been written will raise a
        ValidationError
        """
        raise NotImplementedError("StorageLookup does not support commit_to")
        self.logger.debug2("persist storage root to data store")
        if self._trie_nodes_batch is None:
            raise ValidationError(
                "It is invalid to commit an account's storage if it has no pending "
                "changes. Always check storage_lookup.has_changed_root before "
                "attempting to commit. Write tries on stack = "
                f"{len(self._historical_write_tries)}; Root hash = "
                f"{encode_hex(self._starting_root_hash)}"
            )
        self._trie_nodes_batch.commit_to(db, apply_deletes=False)

        # Mark the trie as having been all written out to the database.
        # It removes the 'dirty' flag and clears out any pending writes.
        self._initialize_to_root_hash(self._write_trie.root_hash)

    def new_trie(self) -> int:
        """
        Switch to an empty trie. Save the old trie, and pending writes, in
        case of a revert.

        :return: index for reviving the previous trie
        """
        raise NotImplementedError("StorageLookup does not support new_trie")
        write_trie = self._get_write_trie()

        # Write the previous trie into a historical stack
        self._historical_write_tries.append(
            PendingWrites(
                write_trie,
                self._trie_nodes_batch,
                self._starting_root_hash,
            )
        )

        new_idx = len(self._historical_write_tries)
        self._starting_root_hash = BLANK_ROOT_HASH
        self._write_trie = None
        self._trie_nodes_batch = None

        return new_idx

    def rollback_trie(self, trie_index: int) -> None:
        """
        Revert back to the previous trie, using the index returned by a
        :meth:`~new_trie` call. The index returned by that call returns you
        to the trie in place *before* the call.

        :param trie_index: index for reviving the previous trie
        """
        raise NotImplementedError("StorageLookup does not support rollback_trie")

        if trie_index >= len(self._historical_write_tries):
            raise ValidationError(
                f"Trying to roll back a delete to index {trie_index}, but there are "
                f"only {len(self._historical_write_tries)} indices available."
            )

        (
            self._write_trie,
            self._trie_nodes_batch,
            self._starting_root_hash,
        ) = self._historical_write_tries[trie_index]

        # Cannot roll forward after a rollback, so remove created/ignored tries.
        # This also deletes the trie that you just reverted to. It will be re-added
        # to the stack when the next new_trie() is called.
        del self._historical_write_tries[trie_index:]


CLEAR_COUNT_KEY_NAME = b"clear-count"


class AccountStorageDB(AccountStorageDatabaseAPI):
    logger = get_extended_debug_logger("eth.db.storage.AccountStorageDB")

    def __init__(
        self,
        db: AtomicDatabaseAPI,
        w3: Web3,
        block_identifier: BlockIdentifier,
        address: Address
    ) -> None:
        """
        Database entries go through several pipes, like so...

        .. code::

          db -> _storage_lookup -> _storage_cache -> _locked_changes -> _journal_storage

        db is the raw database, we can assume it hits disk when written to.
        Keys are stored as node hashes and rlp-encoded node values.

        _storage_lookup is itself a pair of databases: (BatchDB -> HexaryTrie),
        writes to storage lookup *are* immeditaely applied to a trie, generating
        the appropriate trie nodes and and root hash (via the HexaryTrie). The
        writes are *not* persisted to db, until _storage_lookup is explicitly instructed
        to, via :meth:`StorageLookup.commit_to`

        _storage_cache is a cache tied to the state root of the trie. It
        is important that this cache is checked *after* looking for
        the key in _journal_storage, because the cache is only invalidated
        after a state root change. Otherwise, you will see data since the last
        storage root was calculated.

        _locked_changes is a batch database that includes only those values that are
        un-revertable in the EVM. Currently, that means changes that completed in a
        previous transaction.

        Journaling batches writes at the _journal_storage layer, until persist is
        called. It manages all the checkpointing and rollbacks that happen during
        EVM execution.

        In both _storage_cache and _journal_storage, Keys are set/retrieved as the
        big_endian encoding of the slot integer, and the rlp-encoded value.
        """
        if bytes(address).hex() == '0221c903898e883e8c5a51a793768781cd7bbdcc':
            print(f'Constructing AccountStorageDB for 0x0221c903898e883e8c5a51a793768781cd7bbdcc')
        self._address = address
        self._storage_lookup = StorageLookup(db, w3, block_identifier, address)
        self._storage_cache = CacheDB(self._storage_lookup)
        self._locked_changes = JournalDB(self._storage_cache)
        self._journal_storage = JournalDB(self._locked_changes)
        self._accessed_slots: Set[int] = set()

        # Track how many times we have cleared the storage. This is journaled
        # in lockstep with other storage changes. That way, we can detect if a revert
        # causes use to revert past the previous storage deletion. The clear count is
        # used as an index to find the base trie from before the revert.
        self._clear_count = JournalDB(MemoryDB({CLEAR_COUNT_KEY_NAME: to_bytes(0)}))

    def get(self, slot: int, from_journal: bool = True) -> int:
        self._accessed_slots.add(slot)
        key = int_to_big_endian(slot)
        lookup_db = self._journal_storage if from_journal else self._locked_changes

        try:
            v = lookup_db[key]
            if v == b"":
                ret = 0
            else:
                ret = int.from_bytes(v, 'big', signed=False)
        except KeyError:
            ret = 0

        return ret

    def set(self, slot: int, value: int) -> None:
        key = int_to_big_endian(slot)
        if value:
            self._journal_storage[key] = value.to_bytes(32, 'big', signed=False)
        else:
            try:
                current_val = self._journal_storage[key]
            except KeyError:
                # deleting an empty key has no effect
                return
            else:
                if current_val != b"":
                    # only try to delete the value if it's present
                    del self._journal_storage[key]

    def delete(self) -> None:
        self.logger.debug2(
            "Deleting all storage in account 0x%s",
            self._address.hex(),
        )
        self._journal_storage.clear()
        self._storage_cache.reset_cache()

        # Look up the previous count of how many times the account has been deleted.
        # This can happen multiple times in one block, via CREATE2.
        old_clear_count = to_int(self._clear_count[CLEAR_COUNT_KEY_NAME])
        new_clear_count = old_clear_count + 1

        # NOTE: below is now meaningless (since I deleted the old code that gets new_clear_count)
        # # Gut check that we have incremented correctly
        # if new_clear_count != old_clear_count + 1:
        #     raise ValidationError(
        #         f"Must increase clear count by one on each delete. Instead, went from"
        #         f" {old_clear_count} -> {new_clear_count} in account"
        #         f" 0x{self._address.hex()}"
        #     )

        # Save the new count, ie~ the index used for a future revert.
        self._clear_count[CLEAR_COUNT_KEY_NAME] = to_bytes(new_clear_count)

    def record(self, checkpoint: JournalDBCheckpoint) -> None:
        self._journal_storage.record(checkpoint)
        self._clear_count.record(checkpoint)

    def discard(self, checkpoint: JournalDBCheckpoint) -> None:
        self.logger.debug2("discard checkpoint %r", checkpoint)
        latest_clear_count = to_int(self._clear_count[CLEAR_COUNT_KEY_NAME])

        if self._journal_storage.has_checkpoint(checkpoint):
            self._journal_storage.discard(checkpoint)
            self._clear_count.discard(checkpoint)
        else:
            # if the checkpoint comes before this account started tracking,
            #    then simply reset to the beginning (state persisted in remote db)
            self._journal_storage.reset()
            self._clear_count.reset()
        self._storage_cache.reset_cache()

        reverted_clear_count = to_int(self._clear_count[CLEAR_COUNT_KEY_NAME])

        if reverted_clear_count == latest_clear_count - 1:
            # This revert rewinds past a trie deletion (i.e., account deletion),
            # I think we don't need to do anything here, because we only use the journal
            #  to track storage changes
            # self._storage_lookup.rollback_trie(reverted_clear_count)
            pass
        elif reverted_clear_count == latest_clear_count:
            # No change in the base trie, take no action
            pass
        else:
            # Although CREATE2 permits multiple creates and deletes in a single block,
            #   you can still only revert across a single delete. That's because delete
            #   is only triggered at the end of the transaction.
            raise ValidationError(
                f"This revert has changed the clear count in an invalid way, from"
                f" {latest_clear_count} to {reverted_clear_count}, in"
                f" 0x{self._address.hex()}"
            )

    def commit(self, checkpoint: JournalDBCheckpoint) -> None:
        if self._journal_storage.has_checkpoint(checkpoint):
            self._journal_storage.commit(checkpoint)
            self._clear_count.commit(checkpoint)
        else:
            # if the checkpoint comes before this account started tracking,
            #    then flatten all changes, without persisting
            self._journal_storage.flatten()
            self._clear_count.flatten()

    def lock_changes(self) -> None:
        if self._journal_storage.has_clear():
            self._locked_changes.clear()
        self._journal_storage.persist()

    def make_storage_root(self) -> None:
        self.lock_changes()
        self._locked_changes.persist()

    def _validate_flushed(self) -> None:
        """
        Will raise an exception if there are some changes made since the last persist.
        """
        journal_diff = self._journal_storage.diff()
        if len(journal_diff) > 0:
            raise ValidationError(
                "StorageDB had a dirty journal when it needed to be "
                f"clean: {journal_diff!r}"
            )

    def get_accessed_slots(self) -> FrozenSet[int]:
        return frozenset(self._accessed_slots)

    @property
    def has_changed_root(self) -> bool:
        return self._storage_lookup.has_changed_root

    def get_changed_root(self) -> Hash32:
        return self._storage_lookup.get_changed_root()

    def persist(self, db: DatabaseAPI) -> None:
        self._validate_flushed()
        if self._storage_lookup.has_changed_root:
            self._storage_lookup.commit_to(db)



from typing import (
    Callable,
    Tuple,
)

from eth_utils import (
    big_endian_to_int,
    int_to_big_endian,
)
from eth_utils.toolz import (
    curry,
)

from eth import (
    constants,
)
from eth._utils.numeric import (
    get_highest_bit_index,
)
from eth._utils.padding import (
    pad32r,
    zpad_left,
    zpad_right,
)
from eth.abc import (
    ComputationAPI,
)


def _compute_adjusted_exponent_length(
    exponent_length: int, first_32_exponent_bytes: bytes
) -> int:
    exponent = big_endian_to_int(first_32_exponent_bytes)

    if exponent_length <= 32 and exponent == 0:
        return 0
    elif exponent_length <= 32:
        return get_highest_bit_index(exponent)
    else:
        first_32_bytes_as_int = big_endian_to_int(first_32_exponent_bytes)
        return 8 * (exponent_length - 32) + get_highest_bit_index(first_32_bytes_as_int)


def _compute_complexity(length: int) -> int:
    if length <= 64:
        return length**2
    elif length <= 1024:
        return length**2 // 4 + 96 * length - 3072
    else:
        return length**2 // 16 + 480 * length - 199680


def extract_lengths(data: bytes) -> Tuple[int, int, int]:
    # extract argument lengths
    base_length_bytes = pad32r(data[:32])
    base_length = big_endian_to_int(base_length_bytes)

    exponent_length_bytes = pad32r(data[32:64])
    exponent_length = big_endian_to_int(exponent_length_bytes)

    modulus_length_bytes = pad32r(data[64:96])
    modulus_length = big_endian_to_int(modulus_length_bytes)

    return base_length, exponent_length, modulus_length


def _compute_modexp_gas_fee_eip_198(data: bytes) -> int:
    base_length, exponent_length, modulus_length = extract_lengths(data)

    first_32_exponent_bytes = zpad_right(
        data[96 + base_length : 96 + base_length + exponent_length],
        to_size=min(exponent_length, 32),
    )[:32]
    adjusted_exponent_length = _compute_adjusted_exponent_length(
        exponent_length,
        first_32_exponent_bytes,
    )
    complexity = _compute_complexity(max(modulus_length, base_length))

    gas_fee = (
        complexity
        * max(adjusted_exponent_length, 1)
        // constants.GAS_MOD_EXP_QUADRATIC_DENOMINATOR
    )
    return gas_fee


def _modexp(data: bytes) -> int:
    base_length, exponent_length, modulus_length = extract_lengths(data)

    if base_length == 0:
        return 0
    elif modulus_length == 0:
        return 0

    # compute start:end indexes
    base_end_idx = 96 + base_length
    exponent_end_idx = base_end_idx + exponent_length
    modulus_end_dx = exponent_end_idx + modulus_length

    # extract arguments
    modulus_bytes = zpad_right(
        data[exponent_end_idx:modulus_end_dx],
        to_size=modulus_length,
    )
    modulus = big_endian_to_int(modulus_bytes)
    if modulus == 0:
        return 0

    base_bytes = zpad_right(data[96:base_end_idx], to_size=base_length)
    base = big_endian_to_int(base_bytes)

    exponent_bytes = zpad_right(
        data[base_end_idx:exponent_end_idx],
        to_size=exponent_length,
    )
    exponent = big_endian_to_int(exponent_bytes)

    result = pow(base, exponent, modulus)

    return result


@curry
def modexp(
    computation: ComputationAPI,
    gas_calculator: Callable[[bytes], int] = _compute_modexp_gas_fee_eip_198,
) -> ComputationAPI:
    """
    https://github.com/ethereum/EIPs/pull/198
    """
    data = computation.msg.data_as_bytes

    gas_fee = gas_calculator(data)
    computation.consume_gas(gas_fee, reason="MODEXP Precompile")

    result = _modexp(data)

    _, _, modulus_length = extract_lengths(data)

    # Modulo 0 is undefined, return zero
    # https://math.stackexchange.com/questions/516251/why-is-n-mod-0-undefined
    result_bytes = (
        b""
        if modulus_length == 0
        else zpad_left(int_to_big_endian(result), to_size=modulus_length)
    )

    computation.output = result_bytes
    return computation



import rlp
from rlp.sedes import (
    BigEndianInt,
    Binary,
)

address = Binary.fixed_length(20, allow_empty=True)
hash32 = Binary.fixed_length(32)
uint32 = BigEndianInt(32)
uint256 = BigEndianInt(256)
trie_root = Binary.fixed_length(32, allow_empty=True)
chain_gaps = rlp.sedes.List(
    (
        rlp.sedes.CountableList(rlp.sedes.List((uint32, uint32))),
        uint32,
    )
)



import logging

from eth_utils import (
    ValidationError,
)

from eth.abc import (
    DatabaseAPI,
)
from eth.db.backends.base import (
    BaseDB,
)
from eth.db.diff import (
    DBDiff,
    DBDiffTracker,
    DiffMissingError,
)


class BatchDB(BaseDB):
    """
    A wrapper of basic DB objects with uncommitted DB changes stored in local cache,
    which represents as a dictionary of database keys and values.
    This class should be usable as a context manager, the changes either all fail or all
    succeed. Upon exiting the context, it writes all of the key value pairs from the
    cache into the underlying database. If any error occurred before committing phase,
    we would not apply commits at all.
    """

    logger = logging.getLogger("eth.db.BatchDB")

    wrapped_db: DatabaseAPI = None
    _track_diff: DBDiffTracker = None

    def __init__(
        self, wrapped_db: DatabaseAPI, read_through_deletes: bool = False
    ) -> None:
        self.wrapped_db = wrapped_db
        self._track_diff = DBDiffTracker()
        self._read_through_deletes = read_through_deletes

    def __enter__(self) -> "BatchDB":
        return self

    def __exit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
        # commit all the changes from local cache to underlying db
        if exc_type is None:
            self.commit()
        else:
            self.clear()
            self.logger.exception("Unexpected error occurred during batch update")

    def clear(self) -> None:
        self._track_diff = DBDiffTracker()

    def commit(self, apply_deletes: bool = True) -> None:
        self.commit_to(self.wrapped_db, apply_deletes)

    def commit_to(self, target_db: DatabaseAPI, apply_deletes: bool = True) -> None:
        if apply_deletes and self._read_through_deletes:
            raise ValidationError(
                "BatchDB should never apply deletes when reading through deletes"
            )
        diff = self.diff()
        diff.apply_to(target_db, apply_deletes)
        self.clear()

    def _exists(self, key: bytes) -> bool:
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def __getitem__(self, key: bytes) -> bytes:
        try:
            value = self._track_diff[key]
        except DiffMissingError as missing:
            if missing.is_deleted and not self._read_through_deletes:
                raise KeyError(key)
            else:
                return self.wrapped_db[key]
        else:
            return value

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self._track_diff[key] = value

    def __delitem__(self, key: bytes) -> None:
        if key not in self:
            raise KeyError(key)
        del self._track_diff[key]

    def diff(self) -> DBDiff:
        return self._track_diff.diff()



from contextlib import (
    contextmanager,
)
import logging
from typing import (
    Iterator,
)

from eth_utils import (
    ValidationError,
)

from eth.abc import (
    AtomicWriteBatchAPI,
    DatabaseAPI,
)
from eth.db.backends.base import (
    BaseAtomicDB,
    BaseDB,
)
from eth.db.backends.memory import (
    MemoryDB,
)
from eth.db.diff import (
    DBDiff,
    DBDiffTracker,
    DiffMissingError,
)


class AtomicDB(BaseAtomicDB):
    logger = logging.getLogger("eth.db.AtomicDB")

    wrapped_db: DatabaseAPI = None
    _track_diff: DBDiffTracker = None

    def __init__(self, wrapped_db: DatabaseAPI = None) -> None:
        if wrapped_db is None:
            self.wrapped_db = MemoryDB()
        else:
            self.wrapped_db = wrapped_db

    def __getitem__(self, key: bytes) -> bytes:
        return self.wrapped_db[key]

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.wrapped_db[key] = value

    def __delitem__(self, key: bytes) -> None:
        del self.wrapped_db[key]

    def _exists(self, key: bytes) -> bool:
        return key in self.wrapped_db

    @contextmanager
    def atomic_batch(self) -> Iterator[AtomicWriteBatchAPI]:
        with AtomicDBWriteBatch._commit_unless_raises(self) as readable_batch:
            yield readable_batch


class AtomicDBWriteBatch(BaseDB, AtomicWriteBatchAPI):
    """
    This is returned by a BaseAtomicDB during an atomic_batch, to provide a temporary
    view of the database, before commit.
    """

    logger = logging.getLogger("eth.db.AtomicDBWriteBatch")

    _write_target_db: DatabaseAPI = None
    _track_diff: DBDiffTracker = None

    def __init__(self, write_target_db: DatabaseAPI) -> None:
        self._write_target_db = write_target_db
        self._track_diff = DBDiffTracker()

    def __getitem__(self, key: bytes) -> bytes:
        if self._track_diff is None:
            raise ValidationError("Cannot get data from a write batch, out of context")

        try:
            value = self._track_diff[key]
        except DiffMissingError as missing:
            if missing.is_deleted:
                raise KeyError(key)
            else:
                return self._write_target_db[key]
        else:
            return value

    def __setitem__(self, key: bytes, value: bytes) -> None:
        if self._track_diff is None:
            raise ValidationError("Cannot set data from a write batch, out of context")

        self._track_diff[key] = value

    def __delitem__(self, key: bytes) -> None:
        if self._track_diff is None:
            raise ValidationError(
                "Cannot delete data from a write batch, out of context"
            )

        if key not in self:
            raise KeyError(key)
        del self._track_diff[key]

    def _diff(self) -> DBDiff:
        return self._track_diff.diff()

    def _commit(self) -> None:
        self._diff().apply_to(self._write_target_db, apply_deletes=True)

    def _exists(self, key: bytes) -> bool:
        if self._track_diff is None:
            raise ValidationError(
                "Cannot test data existance from a write batch, out of context"
            )

        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    @classmethod
    @contextmanager
    def _commit_unless_raises(
        cls, write_target_db: DatabaseAPI
    ) -> Iterator[AtomicWriteBatchAPI]:
        """
        Commit all writes inside the context, unless an exception was raised.

        Although this is technically an external API, it (and this whole class) is only
        intended to be used by AtomicDB.
        """
        readable_write_batch: AtomicDBWriteBatch = cls(write_target_db)
        try:
            yield readable_write_batch
        except Exception:
            cls.logger.exception(
                "Unexpected error in atomic db write, dropped partial writes: %r",
                readable_write_batch._diff(),
            )
            raise
        else:
            readable_write_batch._commit()
        finally:
            # force a shutdown of this batch, to prevent out-of-context usage
            readable_write_batch._track_diff = None
            readable_write_batch._write_target_db = None



from typing import (
    Type,
)

from eth_utils.toolz import (
    curry,
)
from eth_bloom import (
    BloomFilter,
)
from eth_utils import (
    encode_hex,
    ValidationError,
)

from eth.abc import (
    BlockAPI,
    BlockHeaderAPI,
    ComputationAPI,
    ReceiptAPI,
    SignedTransactionAPI,
    StateAPI,
)
from eth.constants import (
    MAX_UNCLE_DEPTH,
)
from eth.validation import (
    validate_lte,
)
from eth.vm.forks.spurious_dragon import SpuriousDragonVM

from .blocks import ByzantiumBlock
from .constants import (
    EIP649_BLOCK_REWARD,
    EIP658_TRANSACTION_STATUS_CODE_FAILURE,
    EIP658_TRANSACTION_STATUS_CODE_SUCCESS,
)
from .headers import (
    create_byzantium_header_from_parent,
    configure_byzantium_header,
    compute_byzantium_difficulty,
)
from .state import ByzantiumState


@curry
def get_uncle_reward(
    block_reward: int, block_number: int, uncle: BlockHeaderAPI
) -> int:
    block_number_delta = block_number - uncle.block_number
    validate_lte(block_number_delta, MAX_UNCLE_DEPTH)
    return (8 - block_number_delta) * block_reward // 8


EIP658_STATUS_CODES = {
    EIP658_TRANSACTION_STATUS_CODE_SUCCESS,
    EIP658_TRANSACTION_STATUS_CODE_FAILURE,
}


class ByzantiumVM(SpuriousDragonVM):
    # fork name
    fork = "byzantium"

    # classes
    block_class: Type[BlockAPI] = ByzantiumBlock
    _state_class: Type[StateAPI] = ByzantiumState

    # Methods
    create_header_from_parent = staticmethod(create_byzantium_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_byzantium_difficulty)  # type: ignore
    configure_header = configure_byzantium_header
    # Separated into two steps due to mypy bug of staticmethod.
    # https://github.com/python/mypy/issues/5530
    get_uncle_reward = get_uncle_reward(EIP649_BLOCK_REWARD)
    get_uncle_reward = staticmethod(get_uncle_reward)

    @classmethod
    def validate_receipt(cls, receipt: ReceiptAPI) -> None:
        super().validate_receipt(receipt)
        if receipt.state_root not in EIP658_STATUS_CODES:
            raise ValidationError(
                "The receipt's `state_root` must be one of "
                f"[{encode_hex(EIP658_TRANSACTION_STATUS_CODE_SUCCESS)}, "
                f"{encode_hex(EIP658_TRANSACTION_STATUS_CODE_FAILURE)}].  Got: "
                f"{encode_hex(receipt.state_root)}"
            )

    @staticmethod
    def get_block_reward() -> int:
        return EIP649_BLOCK_REWARD

    def add_receipt_to_header(
        self, old_header: BlockHeaderAPI, receipt: ReceiptAPI
    ) -> BlockHeaderAPI:
        # Skip merkelizing the account data and persisting it to disk on every
        # transaction. Starting in Byzantium, this is no longer necessary,
        # because the state root isn't in the receipt anymore.
        return old_header.copy(
            bloom=int(BloomFilter(old_header.bloom) | receipt.bloom),
            gas_used=receipt.gas_used,
        )

    @classmethod
    def make_receipt(
        cls,
        base_header: BlockHeaderAPI,
        transaction: SignedTransactionAPI,
        computation: ComputationAPI,
        state: StateAPI,
    ) -> ReceiptAPI:
        gas_used = base_header.gas_used + cls.finalize_gas_used(
            transaction, computation
        )

        if computation.is_error:
            status_code = EIP658_TRANSACTION_STATUS_CODE_FAILURE
        else:
            status_code = EIP658_TRANSACTION_STATUS_CODE_SUCCESS

        return transaction.make_receipt(
            status_code, gas_used, computation.get_log_entries()
        )



import copy

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.byzantium.opcodes import (
    BYZANTIUM_OPCODES,
    ensure_no_static,
)
from eth.vm.forks.constantinople.constants import (
    GAS_EXTCODEHASH_EIP1052,
)
from eth.vm.forks.constantinople.storage import (
    sstore_eip1283,
)
from eth.vm.logic import (
    arithmetic,
    context,
    system,
)
from eth.vm.opcode import (
    as_opcode,
)

UPDATED_OPCODES = {
    opcode_values.SHL: as_opcode(
        logic_fn=arithmetic.shl,
        mnemonic=mnemonics.SHL,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SHR: as_opcode(
        logic_fn=arithmetic.shr,
        mnemonic=mnemonics.SHR,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SAR: as_opcode(
        logic_fn=arithmetic.sar,
        mnemonic=mnemonics.SAR,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.EXTCODEHASH: as_opcode(
        logic_fn=context.extcodehash,
        mnemonic=mnemonics.EXTCODEHASH,
        gas_cost=GAS_EXTCODEHASH_EIP1052,
    ),
    opcode_values.CREATE2: system.Create2.configure(
        __name__="opcode:CREATE2",
        mnemonic=mnemonics.CREATE2,
        gas_cost=constants.GAS_CREATE,
    )(),
    opcode_values.SSTORE: as_opcode(
        logic_fn=ensure_no_static(sstore_eip1283),
        mnemonic=mnemonics.SSTORE,
        gas_cost=constants.GAS_NULL,
    ),
}

CONSTANTINOPLE_OPCODES = merge(
    copy.deepcopy(BYZANTIUM_OPCODES),
    UPDATED_OPCODES,
)



from typing import Type

from eth.abc import StateAPI
from eth.vm.forks.homestead import HomesteadVM

from .state import TangerineWhistleState


class TangerineWhistleVM(HomesteadVM):
    # fork name
    fork: str = "tangerine-whistle"  # noqa

    # classes
    _state_class: Type[StateAPI] = TangerineWhistleState

    # Don't bother with any DAO logic in Tangerine VM or later
    # This is how we skip DAO logic on Ropsten, for example
    support_dao_fork = False



from eth.vm.forks.istanbul.headers import (
    configure_header,
    create_header_from_parent,
)
from eth.vm.forks.petersburg.headers import (
    compute_difficulty,
)

compute_muir_glacier_difficulty = compute_difficulty(9000000)

create_muir_glacier_header_from_parent = create_header_from_parent(
    compute_muir_glacier_difficulty
)
configure_muir_glacier_header = configure_header(
    difficulty_fn=compute_muir_glacier_difficulty
)



from eth.vm.forks.byzantium.computation import (
    BYZANTIUM_PRECOMPILES,
    ByzantiumComputation,
)

from .opcodes import (
    PETERSBURG_OPCODES,
)

PETERSBURG_PRECOMPILES = BYZANTIUM_PRECOMPILES


class PetersburgComputation(ByzantiumComputation):
    """
    A class for all execution *message* computations in the ``Petersburg`` fork.
    Inherits from
    :class:`~eth.vm.forks.byzantium.computation.ByzantiumComputation`
    """

    # Override
    opcodes = PETERSBURG_OPCODES
    _precompiles = PETERSBURG_PRECOMPILES



from eth import (
    constants,
)
from eth._utils.numeric import (
    signed_to_unsigned,
    unsigned_to_signed,
)
from eth.abc import (
    ComputationAPI,
)


def lt(computation: ComputationAPI) -> None:
    """
    Lesser Comparison
    """
    left, right = computation.stack_pop_ints(2)

    if left < right:
        result = 1
    else:
        result = 0

    computation.stack_push_int(result)


def gt(computation: ComputationAPI) -> None:
    """
    Greater Comparison
    """
    left, right = computation.stack_pop_ints(2)

    if left > right:
        result = 1
    else:
        result = 0

    computation.stack_push_int(result)


def slt(computation: ComputationAPI) -> None:
    """
    Signed Lesser Comparison
    """
    left, right = map(
        unsigned_to_signed,
        computation.stack_pop_ints(2),
    )

    if left < right:
        result = 1
    else:
        result = 0

    computation.stack_push_int(signed_to_unsigned(result))


def sgt(computation: ComputationAPI) -> None:
    """
    Signed Greater Comparison
    """
    left, right = map(
        unsigned_to_signed,
        computation.stack_pop_ints(2),
    )

    if left > right:
        result = 1
    else:
        result = 0

    computation.stack_push_int(signed_to_unsigned(result))


def eq(computation: ComputationAPI) -> None:
    """
    Equality
    """
    left, right = computation.stack_pop_ints(2)

    if left == right:
        result = 1
    else:
        result = 0

    computation.stack_push_int(result)


def iszero(computation: ComputationAPI) -> None:
    """
    Not
    """
    value = computation.stack_pop1_int()

    if value == 0:
        result = 1
    else:
        result = 0

    computation.stack_push_int(result)


def and_op(computation: ComputationAPI) -> None:
    """
    Bitwise And
    """
    left, right = computation.stack_pop_ints(2)

    result = left & right

    computation.stack_push_int(result)


def or_op(computation: ComputationAPI) -> None:
    """
    Bitwise Or
    """
    left, right = computation.stack_pop_ints(2)

    result = left | right

    computation.stack_push_int(result)


def xor(computation: ComputationAPI) -> None:
    """
    Bitwise XOr
    """
    left, right = computation.stack_pop_ints(2)

    result = left ^ right

    computation.stack_push_int(result)


def not_op(computation: ComputationAPI) -> None:
    """
    Not
    """
    value = computation.stack_pop1_int()

    result = constants.UINT_256_MAX - value

    computation.stack_push_int(result)


def byte_op(computation: ComputationAPI) -> None:
    """
    Bitwise And
    """
    position, value = computation.stack_pop_ints(2)

    if position >= 32:
        result = 0
    else:
        result = (value // pow(256, 31 - position)) % 256

    computation.stack_push_int(result)



from typing import (
    Any,
    Callable,
    List,
    Optional,
)

from eth_utils import (
    ValidationError,
)
from toolz.functoolz import (
    curry,
)

from eth._utils.headers import (
    compute_gas_limit,
    fill_header_params_from_parent,
    new_timestamp_from_parent,
)
from eth.abc import (
    BlockHeaderAPI,
    BlockHeaderSedesAPI,
)
from eth.constants import (
    GENESIS_GAS_LIMIT,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.vm.forks.berlin.headers import (
    configure_header,
)
from eth.vm.forks.muir_glacier.headers import (
    compute_difficulty,
)

from .blocks import (
    LondonBlockHeader,
)
from .constants import (
    BASE_FEE_MAX_CHANGE_DENOMINATOR,
    ELASTICITY_MULTIPLIER,
    INITIAL_BASE_FEE,
)


def calculate_expected_base_fee_per_gas(parent_header: BlockHeaderAPI) -> int:
    if parent_header is None:
        # Parent is empty when making the genesis header
        return INITIAL_BASE_FEE
    else:
        try:
            parent_base_fee_per_gas = parent_header.base_fee_per_gas
        except AttributeError:
            # Parent is a non-London header
            return INITIAL_BASE_FEE

    # Parent *is* a London header
    parent_gas_target = parent_header.gas_limit // ELASTICITY_MULTIPLIER
    parent_gas_used = parent_header.gas_used

    if parent_gas_used == parent_gas_target:
        return parent_base_fee_per_gas

    elif parent_gas_used > parent_gas_target:
        gas_used_delta = parent_gas_used - parent_gas_target
        overburnt_wei = parent_base_fee_per_gas * gas_used_delta
        base_fee_per_gas_delta = max(
            overburnt_wei // parent_gas_target // BASE_FEE_MAX_CHANGE_DENOMINATOR,
            1,
        )
        return parent_base_fee_per_gas + base_fee_per_gas_delta

    else:
        gas_used_delta = parent_gas_target - parent_gas_used
        underburnt_wei = parent_base_fee_per_gas * gas_used_delta
        base_fee_per_gas_delta = (
            underburnt_wei // parent_gas_target // BASE_FEE_MAX_CHANGE_DENOMINATOR
        )
        return max(parent_base_fee_per_gas - base_fee_per_gas_delta, 0)


@curry
def create_london_header_from_parent(
    difficulty_fn: Callable[[BlockHeaderAPI, int], int],
    parent_header: Optional[BlockHeaderAPI],
    **header_params: Any,
) -> BlockHeaderAPI:
    if "gas_limit" not in header_params:
        if parent_header is not None and not hasattr(parent_header, "base_fee_per_gas"):
            # If the previous block was not a London block,
            #   double the gas limit, so the new target is the old gas limit
            header_params["gas_limit"] = parent_header.gas_limit * ELASTICITY_MULTIPLIER
        else:
            # frontier rules
            header_params["gas_limit"] = compute_gas_limit(
                parent_header,
                genesis_gas_limit=GENESIS_GAS_LIMIT,
            )

    # byzantium
    if "timestamp" not in header_params:
        header_params["timestamp"] = new_timestamp_from_parent(parent_header)

    if "difficulty" not in header_params:
        if parent_header is None:
            raise ValueError(
                "Must set difficulty when creating a new genesis header (no parent)."
                " Consider 1 for easy mining or eth.constants.GENESIS_DIFFICULTY for"
                " consistency."
            )
        else:
            header_params["difficulty"] = difficulty_fn(
                parent_header,
                header_params["timestamp"],
            )

    # The general fill function doesn't recognize this custom field, so remove it
    configured_fee_per_gas = header_params.pop("base_fee_per_gas", None)

    all_fields = fill_header_params_from_parent(parent_header, **header_params)

    calculated_fee_per_gas = calculate_expected_base_fee_per_gas(parent_header)
    if configured_fee_per_gas is None:
        all_fields["base_fee_per_gas"] = calculated_fee_per_gas
    else:
        # Must not configure an invalid base fee. So verify that either:
        #   1. This is the genesis header, or
        #   2. The configured value matches the calculated value from the parent
        if parent_header is None or configured_fee_per_gas == calculated_fee_per_gas:
            all_fields["base_fee_per_gas"] = configured_fee_per_gas
        else:
            raise ValidationError(
                f"Cannot select an invalid base_fee_per_gas of:"
                f" {configured_fee_per_gas}, expected: {calculated_fee_per_gas}"
            )

    new_header = LondonBlockHeader(**all_fields)  # type:ignore
    return new_header


compute_london_difficulty = compute_difficulty(9700000)
configure_london_header = configure_header(difficulty_fn=compute_london_difficulty)


class LondonBackwardsHeader(BlockHeaderSedesAPI):
    """
    An rlp sedes class for block headers.

    It can serialize and deserialize *both* London and pre-London headers.
    """

    @classmethod
    def serialize(cls, obj: BlockHeaderAPI) -> List[bytes]:
        if isinstance(obj, LondonBlockHeader):
            return LondonBlockHeader.serialize(obj)
        else:
            return BlockHeader.serialize(obj)

    @classmethod
    def deserialize(cls, encoded: List[bytes]) -> BlockHeaderAPI:
        num_fields = len(encoded)
        if num_fields == 16:
            return LondonBlockHeader.deserialize(encoded)
        elif num_fields == 15:
            return BlockHeader.deserialize(encoded)
        else:
            raise ValueError(
                "London & earlier can only handle headers of 15 or 16 fields. "
                f"Got {num_fields} in {encoded!r}"
            )



from eth_utils.toolz import (
    curry,
)

from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    OutOfGas,
)
from eth.vm.forks.constantinople.storage import (
    GAS_SCHEDULE_EIP1283,
)
from eth.vm.forks.istanbul import (
    constants,
)
from eth.vm.logic.storage import (
    NetSStoreGasSchedule,
    net_sstore,
)

GAS_SCHEDULE_EIP2200 = GAS_SCHEDULE_EIP1283._replace(
    sload_gas=constants.GAS_SLOAD_EIP1884,
)


@curry
def sstore_eip2200_generic(
    gas_schedule: NetSStoreGasSchedule,
    computation: ComputationAPI,
) -> int:
    gas_remaining = computation.get_gas_remaining()
    if gas_remaining <= 2300:
        raise OutOfGas(
            "Net-metered SSTORE always fails below 2300 gas, per EIP-2200",
            gas_remaining,
        )
    else:
        return net_sstore(gas_schedule, computation)


sstore_eip2200 = sstore_eip2200_generic(GAS_SCHEDULE_EIP2200)



#
# Stop and Arithmetic
#
STOP = 0x00
ADD = 0x01
MUL = 0x02
SUB = 0x03
DIV = 0x04
SDIV = 0x05
MOD = 0x06
SMOD = 0x07
ADDMOD = 0x08
MULMOD = 0x09
EXP = 0x0A
SIGNEXTEND = 0x0B


#
# Comparison and Bitwise Logic
#
LT = 0x10
GT = 0x11
SLT = 0x12
SGT = 0x13
EQ = 0x14
ISZERO = 0x15
AND = 0x16
OR = 0x17
XOR = 0x18
NOT = 0x19
BYTE = 0x1A
SHL = 0x1B
SHR = 0x1C
SAR = 0x1D


#
# Sha3
#
SHA3 = 0x20


#
# Environment Information
#
ADDRESS = 0x30
BALANCE = 0x31
ORIGIN = 0x32
CALLER = 0x33
CALLVALUE = 0x34
CALLDATALOAD = 0x35
CALLDATASIZE = 0x36
CALLDATACOPY = 0x37
CODESIZE = 0x38
CODECOPY = 0x39
GASPRICE = 0x3A
EXTCODESIZE = 0x3B
EXTCODECOPY = 0x3C
RETURNDATASIZE = 0x3D
RETURNDATACOPY = 0x3E
EXTCODEHASH = 0x3F

# These opcodes seem to belong in the environment block,
# but we are out of opcode space in 0x3*
CHAINID = 0x46
SELFBALANCE = 0x47
BASEFEE = 0x48

#
# Block Information
#
BLOCKHASH = 0x40
COINBASE = 0x41
TIMESTAMP = 0x42
NUMBER = 0x43
DIFFICULTY = 0x44
PREVRANDAO = 0x44  # EIP-4399: supplant DIFFICULTY with PREVRANDAO
GASLIMIT = 0x45


#
# Stack, Memory, Storage and Flow Operations
#
POP = 0x50

MLOAD = 0x51
MSTORE = 0x52
MSTORE8 = 0x53
SLOAD = 0x54
SSTORE = 0x55
JUMP = 0x56
JUMPI = 0x57
PC = 0x58
MSIZE = 0x59
GAS = 0x5A
JUMPDEST = 0x5B


#
# Push Operations
#
PUSH0 = 0x5F
PUSH1 = 0x60
PUSH2 = 0x61
PUSH3 = 0x62
PUSH4 = 0x63
PUSH5 = 0x64
PUSH6 = 0x65
PUSH7 = 0x66
PUSH8 = 0x67
PUSH9 = 0x68
PUSH10 = 0x69
PUSH11 = 0x6A
PUSH12 = 0x6B
PUSH13 = 0x6C
PUSH14 = 0x6D
PUSH15 = 0x6E
PUSH16 = 0x6F
PUSH17 = 0x70
PUSH18 = 0x71
PUSH19 = 0x72
PUSH20 = 0x73
PUSH21 = 0x74
PUSH22 = 0x75
PUSH23 = 0x76
PUSH24 = 0x77
PUSH25 = 0x78
PUSH26 = 0x79
PUSH27 = 0x7A
PUSH28 = 0x7B
PUSH29 = 0x7C
PUSH30 = 0x7D
PUSH31 = 0x7E
PUSH32 = 0x7F


#
# Duplicate Operations
#
DUP1 = 0x80
DUP2 = 0x81
DUP3 = 0x82
DUP4 = 0x83
DUP5 = 0x84
DUP6 = 0x85
DUP7 = 0x86
DUP8 = 0x87
DUP9 = 0x88
DUP10 = 0x89
DUP11 = 0x8A
DUP12 = 0x8B
DUP13 = 0x8C
DUP14 = 0x8D
DUP15 = 0x8E
DUP16 = 0x8F


#
# Exchange Operations
#
SWAP1 = 0x90
SWAP2 = 0x91
SWAP3 = 0x92
SWAP4 = 0x93
SWAP5 = 0x94
SWAP6 = 0x95
SWAP7 = 0x96
SWAP8 = 0x97
SWAP9 = 0x98
SWAP10 = 0x99
SWAP11 = 0x9A
SWAP12 = 0x9B
SWAP13 = 0x9C
SWAP14 = 0x9D
SWAP15 = 0x9E
SWAP16 = 0x9F


#
# Logging
#
LOG0 = 0xA0
LOG1 = 0xA1
LOG2 = 0xA2
LOG3 = 0xA3
LOG4 = 0xA4


#
# System
#
CREATE = 0xF0
CALL = 0xF1
CALLCODE = 0xF2
RETURN = 0xF3
DELEGATECALL = 0xF4
CREATE2 = 0xF5
STATICCALL = 0xFA
REVERT = 0xFD
SELFDESTRUCT = 0xFF



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)


extras_require = {
    "eth": [
        "cached-property>=1.5.1,<2",
        "eth-bloom>=1.0.3",
        "eth-keys>=0.4.0,<0.5.0",
        "eth-typing>=3.3.0,<4.0.0",
        "eth-utils>=2.0.0,<3.0.0",
        "lru-dict>=1.1.6",
        "mypy_extensions>=1.0.0",
        "py-ecc>=1.4.7,<7.0.0",
        "pyethash>=0.1.27,<1.0.0",
        "rlp>=3,<4",
        "trie>=2.0.0,<3",
    ],
    # The eth-extra sections is for libraries that the evm does not
    # explicitly need to function and hence should not depend on.
    # Installing these libraries may make the evm perform better than
    # using the default fallbacks though.
    "eth-extra": [
        "blake2b-py>=0.1.4,<0.2",
        "coincurve>=13.0.0,<14.0.0",
        "eth-hash[pysha3];implementation_name=='cpython'",
        "eth-hash[pycryptodome];implementation_name=='pypy'",
        "plyvel>=1.2.0,<2",
    ],
    "test": [
        "factory-boy==2.11.1",
        "hypothesis>=5,<6",
        "pexpect>=4.6, <5",
        "pytest>=6.2.4,<7",
        "pytest-asyncio>=0.10.0,<0.11",
        "pytest-cov==2.5.1",
        "pytest-timeout>=1.4.2,<2",
        "pytest-watch>=4.1.0,<5",
        "pytest-xdist==2.3.0",
        "importlib-metadata<5.0;python_version<'3.8'",
    ],
    "lint": [
        "flake8==6.0.0",  # flake8 claims semver but adds new warnings at minor releases, leave it pinned.
        "flake8-bugbear==23.3.23",  # flake8-bugbear does not follow semver, leave it pinned.
        "isort>=5.10.1",
        "mypy==1.4.0",  # mypy does not follow semver, leave it pinned.
        "pydocstyle>=6.0.0",
        "black>=23",
        "types-setuptools",
        "importlib-metadata<5.0;python_version<'3.8'",
    ],
    "benchmark": [
        "termcolor>=1.1.0,<2.0.0",
        "web3>=4.1.0,<5.0.0",
    ],
    "docs": [
        "py-evm>=0.2.0-a.14",
        # We need to have pysha for autodoc to be able to extract API docs
        "pysha3>=1.0.0,<2.0.0",
        "Sphinx>=1.5.5,<2",
        "jinja2>=3.0.0,<3.1.0",  # jinja2<3.0 or >=3.1.0 cause doc build failures.
        "sphinx_rtd_theme>=0.1.9",
        "sphinxcontrib-asyncio>=0.2.0,<0.4",
        "towncrier>=21,<22",
    ],
    "dev": [
        "bumpversion>=0.5.3,<1",
        "wheel",
        "setuptools>=36.2.0",
        # Fixing this dependency due to: requests 2.20.1 has requirement
        # idna<2.8,>=2.5, but you'll have idna 2.8 which is incompatible.
        "idna==2.7",
        # idna 2.7 is not supported by requests 2.18
        "requests>=2.20,<3",
        "tox>=4.0.0",
        "twine",
    ],
}


extras_require["dev"] = (
    extras_require["dev"]
    + extras_require["eth"]
    + extras_require["test"]
    + extras_require["lint"]
    + extras_require["docs"]
)

install_requires = extras_require["eth"]

with open("README.md") as readme_file:
    long_description = readme_file.read()

setup(
    name="ethpwn-py-evm",
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version="0.7.0-alpha.4",
    description="Python implementation of the Ethereum Virtual Machine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ethereum Foundation",
    author_email="piper@pipermerriam.com",
    url="https://github.com/ethereum/py-evm",
    include_package_data=True,
    py_modules=["eth"],
    install_requires=install_requires,
    extras_require=extras_require,
    license="MIT",
    zip_safe=False,
    keywords="ethereum blockchain evm",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"eth": ["py.typed"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



from .base import (  # noqa: F401
    Chain,
)
from .mainnet import (  # noqa: F401
    MainnetChain,
)
from .ropsten import (  # noqa: F401
    RopstenChain,
)
from .tester import (  # noqa: F401
    MainnetTesterChain,
)



import pkg_resources
import sys

from eth.chains import (  # noqa: F401
    Chain,
    MainnetChain,
    MainnetTesterChain,
    RopstenChain,
)

#
#  Ensure we can reach 1024 frames of recursion
#
EVM_RECURSION_LIMIT = 1024 * 12
sys.setrecursionlimit(max(EVM_RECURSION_LIMIT, sys.getrecursionlimit()))


__version__ = pkg_resources.get_distribution("ethpwn-py-evm").version



import collections
from itertools import (
    count,
)
from typing import (
    Callable,
    Dict,
    List,
    Set,
    Union,
    cast,
)

from eth_utils import (
    ValidationError,
)
from eth_utils.toolz import (
    first,
    nth,
)

from eth.abc import (
    DatabaseAPI,
)
from eth.typing import (
    JournalDBCheckpoint,
)

from .backends.base import (
    BaseDB,
)
from .diff import (
    DBDiff,
    DBDiffTracker,
)


class DeletedEntry:
    pass


# Track two different kinds of deletion:

# 1. key in wrapped
# 2. key modified in journal
# 3. key deleted
DELETE_WRAPPED = DeletedEntry()

# 1. key not in wrapped
# 2. key created in journal
# 3. key deleted
REVERT_TO_WRAPPED = DeletedEntry()

ChangesetValue = Union[bytes, DeletedEntry]
ChangesetDict = Dict[bytes, ChangesetValue]

get_next_checkpoint = cast(Callable[[], JournalDBCheckpoint], count().__next__)


class Journal(BaseDB):
    """
    A Journal provides a mechanism to track a series of changes to a dict, by inserting
    checkpoints, and committing to them or rolling back to them, and ultimitely
    persisting the final changes.

    Internally, it keeps an ordered list of reversion changesets, used to roll back
    on demand. This is optimized for the most common path: lots of checkpoints and
    commits, and not many discards.

    Checkpoints are referenced by an internally-generated integer.
    This is *not* threadsafe.
    """

    __slots__ = [
        "_journal_data",
        "_clears_at",
        "_current_values",
        "_ignore_wrapped_db",
        "_checkpoint_stack",
    ]

    #
    # This is a high-use class, where we sometimes prefer optimization over
    # readability. It's most important to optimize for record, commit, and persist,
    # which are the most commonly used methods.
    #

    def __init__(self) -> None:
        # If the journal was persisted right now,
        # these would be the current changes to push:
        self._current_values: ChangesetDict = {}

        # contains a mapping from all of the int checkpoints
        # to a dictionary of key:value pairs that are used to rewind from the current
        # values to the given checkpoint
        self._journal_data: collections.OrderedDict[
            JournalDBCheckpoint, ChangesetDict
        ] = collections.OrderedDict()

        # Clears are special operations that enforce that the underlying database and
        # current changes are completely emptied out. Clears are also committable &
        # discardable.
        self._clears_at: Set[JournalDBCheckpoint] = set()

        # If a clear was called, then any missing keys should be treated as missing
        self._ignore_wrapped_db = False

        # To speed up commits, we leave in old recorded checkpoints in
        # self._journal_data, even on commit. Instead of dropping them, we keep a
        # separate list of active checkpoints.
        self._checkpoint_stack: List[JournalDBCheckpoint] = []

    @property
    def root_checkpoint(self) -> JournalDBCheckpoint:
        """
        Returns the starting checkpoint
        """
        return first(self._journal_data.keys())

    @property
    def is_flattened(self) -> bool:
        """
        :return: whether there are any explicitly committed checkpoints
        """
        return len(self._checkpoint_stack) < 2

    @property
    def last_checkpoint(self) -> JournalDBCheckpoint:
        """
        Returns the latest checkpoint
        """
        # last() was iterating through all values, so first(reversed()) gives a 12.5x
        # speedup Interestingly, an attempt to cache this value caused a slowdown.
        return first(reversed(self._journal_data.keys()))

    def has_checkpoint(self, checkpoint: JournalDBCheckpoint) -> bool:
        # another option would be to enforce monotonically-increasing checkpoints,
        # so we can do:
        # checkpoint_idx = bisect_left(self._checkpoint_stack, checkpoint)
        # (then validate against length and value at index)
        return checkpoint in self._checkpoint_stack

    def record_checkpoint(
        self, custom_checkpoint: JournalDBCheckpoint = None
    ) -> JournalDBCheckpoint:
        """
        Creates a new checkpoint. Checkpoints are a sequential int chosen by Journal
        to prevent collisions.
        """
        if custom_checkpoint is not None:
            if custom_checkpoint in self._journal_data:
                raise ValidationError(
                    "Tried to record with an existing checkpoint: "
                    f"{custom_checkpoint!r}"
                )
            else:
                checkpoint = custom_checkpoint
        else:
            checkpoint = get_next_checkpoint()

        self._journal_data[checkpoint] = {}
        self._checkpoint_stack.append(checkpoint)
        return checkpoint

    def discard(self, through_checkpoint_id: JournalDBCheckpoint) -> None:
        while self._checkpoint_stack:
            checkpoint_id = self._checkpoint_stack.pop()
            if checkpoint_id == through_checkpoint_id:
                break
        else:
            # checkpoint not found!
            raise ValidationError(f"No checkpoint {through_checkpoint_id} was found")

        # This might be optimized further by iterating the other direction and
        # ignoring any follow-up rollbacks on the same variable.
        for _ in range(len(self._journal_data)):
            checkpoint_id, rollback_data = self._journal_data.popitem()

            for old_key, old_value in rollback_data.items():
                if old_value is REVERT_TO_WRAPPED:
                    # The current value may not exist, if it was a delete followed by a
                    # clear, so pop it off, or ignore if it is already missing
                    self._current_values.pop(old_key, None)
                elif old_value is DELETE_WRAPPED:
                    self._current_values[old_key] = old_value
                else:
                    self._current_values[old_key] = old_value

            if checkpoint_id in self._clears_at:
                self._clears_at.remove(checkpoint_id)
                self._ignore_wrapped_db = False

            if checkpoint_id == through_checkpoint_id:
                break

        if self._clears_at:
            # if there is still a clear in older locations,
            # then reinitiate the clear flag
            self._ignore_wrapped_db = True

    def clear(self) -> None:
        """
        Treat as if the *underlying* database will also be cleared by some other
        mechanism. We build a special empty reversion changeset just for marking that
        all previous data should be ignored.
        """
        checkpoint = get_next_checkpoint()
        self._journal_data[checkpoint] = self._current_values
        self._current_values = {}
        self._ignore_wrapped_db = True
        self._clears_at.add(checkpoint)

    def has_clear(self, at_checkpoint: JournalDBCheckpoint) -> bool:
        for reversion_changeset_id in reversed(self._journal_data.keys()):
            if reversion_changeset_id in self._clears_at:
                return True
            elif at_checkpoint == reversion_changeset_id:
                return False
        raise ValidationError(f"Checkpoint {at_checkpoint} is not in the journal")

    def commit_checkpoint(self, commit_to: JournalDBCheckpoint) -> ChangesetDict:
        """
        Collapses all changes since the given checkpoint. Can no longer discard to any
        of the checkpoints that followed the given checkpoint.
        """
        # Another option would be to enforce monotonically-increasing changeset ids,
        # so we can do:
        # checkpoint_idx = bisect_left(self._checkpoint_stack, commit_to)
        # (then validate against length and value at index)
        for positions_before_last, checkpoint in enumerate(
            reversed(self._checkpoint_stack)
        ):
            if checkpoint == commit_to:
                checkpoint_idx = -1 - positions_before_last
                break
        else:
            raise ValidationError(f"No checkpoint {commit_to} was found")

        if checkpoint_idx == -1 * len(self._checkpoint_stack):
            raise ValidationError(
                "Should not commit root changeset with commit_changeset, "
                "use pop_all() instead"
            )

        # delete committed checkpoints from the stack
        # (but keep rollbacks for future discards)
        del self._checkpoint_stack[checkpoint_idx:]

        return self._current_values

    def pop_all(self) -> ChangesetDict:
        final_changes = self._current_values
        self._journal_data.clear()
        self._clears_at.clear()
        self._current_values = {}
        self._checkpoint_stack.clear()
        self.record_checkpoint()
        self._ignore_wrapped_db = False
        return final_changes

    def flatten(self) -> None:
        if self.is_flattened:
            return

        checkpoint_after_root = nth(1, self._checkpoint_stack)
        self.commit_checkpoint(checkpoint_after_root)

    #
    # Database API
    #
    def __getitem__(self, key: bytes) -> ChangesetValue:  # type: ignore # Breaks LSP
        """
        For key lookups we need to iterate through the changesets in reverse
        order, returning from the first one in which the key is present.
        """
        # the default result (the value if not in the local values) depends on whether
        # there was a clear
        if self._ignore_wrapped_db:
            default_result = REVERT_TO_WRAPPED
        else:
            default_result = None  # indicate that caller should check wrapped database
        return self._current_values.get(key, default_result)

    def __setitem__(self, key: bytes, value: bytes) -> None:
        # if the value has not been changed since wrapping,
        # then simply revert to original value
        revert_changeset = self._journal_data[self.last_checkpoint]
        if key not in revert_changeset:
            revert_changeset[key] = self._current_values.get(key, REVERT_TO_WRAPPED)
        self._current_values[key] = value

    def _exists(self, key: bytes) -> bool:
        val = self.get(key)
        return val is not None and val not in (REVERT_TO_WRAPPED, DELETE_WRAPPED)

    def __delitem__(self, key: bytes) -> None:
        raise NotImplementedError(
            "You must delete with one of delete_local or delete_wrapped"
        )

    def delete_wrapped(self, key: bytes) -> None:
        revert_changeset = self._journal_data[self.last_checkpoint]
        if key not in revert_changeset:
            revert_changeset[key] = self._current_values.get(key, REVERT_TO_WRAPPED)
        self._current_values[key] = DELETE_WRAPPED

    def delete_local(self, key: bytes) -> None:
        revert_changeset = self._journal_data[self.last_checkpoint]
        if key not in revert_changeset:
            revert_changeset[key] = self._current_values.get(key, REVERT_TO_WRAPPED)
        self._current_values[key] = REVERT_TO_WRAPPED

    def diff(self) -> DBDiff:
        tracker = DBDiffTracker()

        for key, value in self._current_values.items():
            if value is DELETE_WRAPPED:
                del tracker[key]
            elif value is REVERT_TO_WRAPPED:
                pass
            else:
                tracker[key] = value  # type: ignore  # cast(bytes, value)

        return tracker.diff()


class JournalDB(BaseDB):
    """
    A wrapper around the basic DB objects that keeps a journal of all changes.
    Checkpoints can be recorded at any time. You can then commit or roll back
    to those checkpoints.

    Discarding a checkpoint throws away all changes that happened since that
    checkpoint.
    Commiting a checkpoint simply removes the option of reverting back to it
    later.

    Nothing is written to the underlying db until `persist()` is called.

    The added memory footprint for a JournalDB is one key/value stored per
    database key which is changed, at each checkpoint.  Subsequent changes to the same
    key between two checkpoints will not increase the journal size, since we
    do not permit reverting to a place that has no checkpoint.
    """

    __slots__ = ["_wrapped_db", "_journal", "record", "commit"]

    def __init__(self, wrapped_db: DatabaseAPI) -> None:
        self._wrapped_db = wrapped_db
        self._journal = Journal()
        self.record = self._journal.record_checkpoint
        self.commit = self._journal.commit_checkpoint
        self.reset()

    def __getitem__(self, key: bytes) -> bytes:
        val = self._journal[key]
        if val is DELETE_WRAPPED:
            raise KeyError(
                key,
                "item is deleted in JournalDB, and will be deleted from the wrapped DB",
            )
        elif val is REVERT_TO_WRAPPED:
            raise KeyError(
                key,
                "item is deleted in JournalDB, "
                "and is presumed gone from the wrapped DB",
            )
        elif val is None:
            return self._wrapped_db[key]
        else:
            # mypy doesn't allow custom type guards yet so we need to cast here
            # even though we know it can only be `bytes` at this point.
            return cast(bytes, val)

    def __setitem__(self, key: bytes, value: bytes) -> None:
        """
        - replacing an existing value
        - setting a value that does not exist
        """
        self._journal[key] = value

    def _exists(self, key: bytes) -> bool:
        val = self._journal[key]
        if val in (REVERT_TO_WRAPPED, DELETE_WRAPPED):
            return False
        elif val is None:
            return key in self._wrapped_db
        else:
            return True

    def clear(self) -> None:
        """
        Remove all keys. Immediately after a clear, *all* getitem requests will return a
        KeyError. That includes the changes pending persist and any data in the
        underlying database.

        (This action is journaled, like all other actions)

        clear will *not* persist the emptying of all keys in the underlying DB.
        It only prevents any updates (or deletes!) before it from being persisted.

        Any caller that wants to use clear must also make sure that the underlying
        database reflects their desired end state (maybe emptied, maybe not).
        """
        self._journal.clear()

    def has_clear(self) -> bool:
        return self._journal.has_clear(self._journal.root_checkpoint)

    def __delitem__(self, key: bytes) -> None:
        if key in self._wrapped_db:
            self._journal.delete_wrapped(key)
        else:
            if key in self._journal:
                self._journal.delete_local(key)
            else:
                raise KeyError(
                    key, "key could not be deleted in JournalDB, because it was missing"
                )

    #
    # Snapshot API
    #
    def has_checkpoint(self, checkpoint: JournalDBCheckpoint) -> bool:
        return self._journal.has_checkpoint(checkpoint)

    def discard(self, checkpoint: JournalDBCheckpoint) -> None:
        """
        Throws away all journaled data starting at the given checkpoint
        """
        self._journal.discard(checkpoint)

    def _reapply_checkpoint_to_journal(self, journal_data: ChangesetDict) -> None:
        for key, value in journal_data.items():
            if value is DELETE_WRAPPED:
                self._journal.delete_wrapped(key)
            elif value is REVERT_TO_WRAPPED:
                self._journal.delete_local(key)
            else:
                self._journal[key] = cast(bytes, value)

    def persist(self) -> None:
        """
        Persist all changes in underlying db. After all changes have been written the
        JournalDB starts a new recording.
        """
        journal_data = self._journal.pop_all()

        for key, value in journal_data.items():
            try:
                if value is DELETE_WRAPPED:
                    del self._wrapped_db[key]
                elif value is REVERT_TO_WRAPPED:
                    pass
                else:
                    self._wrapped_db[key] = cast(bytes, value)
            except Exception:
                self._reapply_checkpoint_to_journal(journal_data)
                raise

    def flatten(self) -> None:
        """
        Commit everything possible without persisting
        """
        self._journal.flatten()

    def reset(self) -> None:
        """
        Reset the entire journal.
        """
        self._journal.pop_all()

    def diff(self) -> DBDiff:
        """
        Generate a DBDiff of all pending changes.
        These are the changes that would occur if :meth:`persist()` were called.
        """
        return self._journal.diff()



from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.state import BaseState

from .blocks import GrayGlacierBlock
from .headers import (
    compute_gray_glacier_difficulty,
    configure_gray_glacier_header,
    create_gray_glacier_header_from_parent,
)
from .state import GrayGlacierState
from .. import ArrowGlacierVM


class GrayGlacierVM(ArrowGlacierVM):
    # fork name
    fork = "gray-glacier"

    # classes
    block_class: Type[BaseBlock] = GrayGlacierBlock
    _state_class: Type[BaseState] = GrayGlacierState

    # Methods
    create_header_from_parent = staticmethod(  # type: ignore
        create_gray_glacier_header_from_parent(compute_gray_glacier_difficulty)
    )
    compute_difficulty = staticmethod(compute_gray_glacier_difficulty)  # type: ignore
    configure_header = configure_gray_glacier_header



from eth_typing import (
    BlockNumber,
)

GOERLI_CHAIN_ID = 5

# Fork Blocks listed in ascending order

#
# Petersburg
#
PETERSBURG_GOERLI_BLOCK = BlockNumber(0)

#
# Istanbul
#
ISTANBUL_GOERLI_BLOCK = BlockNumber(1561651)

BERLIN_GOERLI_BLOCK = BlockNumber(4460644)



from eth_utils.toolz import (
    curry,
)

from eth import (
    constants,
)
from eth._utils.numeric import (
    ceil8,
    signed_to_unsigned,
    unsigned_to_signed,
)
from eth.abc import (
    ComputationAPI,
)


def add(computation: ComputationAPI) -> None:
    """
    Addition
    """
    left, right = computation.stack_pop_ints(2)

    result = (left + right) & constants.UINT_256_MAX

    computation.stack_push_int(result)


def addmod(computation: ComputationAPI) -> None:
    """
    Modulo Addition
    """
    left, right, mod = computation.stack_pop_ints(3)

    if mod == 0:
        result = 0
    else:
        result = (left + right) % mod

    computation.stack_push_int(result)


def sub(computation: ComputationAPI) -> None:
    """
    Subtraction
    """
    left, right = computation.stack_pop_ints(2)

    result = (left - right) & constants.UINT_256_MAX

    computation.stack_push_int(result)


def mod(computation: ComputationAPI) -> None:
    """
    Modulo
    """
    value, mod = computation.stack_pop_ints(2)

    if mod == 0:
        result = 0
    else:
        result = value % mod

    computation.stack_push_int(result)


def smod(computation: ComputationAPI) -> None:
    """
    Signed Modulo
    """
    value, mod = map(
        unsigned_to_signed,
        computation.stack_pop_ints(2),
    )

    pos_or_neg = -1 if value < 0 else 1

    if mod == 0:
        result = 0
    else:
        result = (abs(value) % abs(mod) * pos_or_neg) & constants.UINT_256_MAX

    computation.stack_push_int(signed_to_unsigned(result))


def mul(computation: ComputationAPI) -> None:
    """
    Multiplication
    """
    left, right = computation.stack_pop_ints(2)

    result = (left * right) & constants.UINT_256_MAX

    computation.stack_push_int(result)


def mulmod(computation: ComputationAPI) -> None:
    """
    Modulo Multiplication
    """
    left, right, mod = computation.stack_pop_ints(3)

    if mod == 0:
        result = 0
    else:
        result = (left * right) % mod
    computation.stack_push_int(result)


def div(computation: ComputationAPI) -> None:
    """
    Division
    """
    numerator, denominator = computation.stack_pop_ints(2)

    if denominator == 0:
        result = 0
    else:
        result = (numerator // denominator) & constants.UINT_256_MAX

    computation.stack_push_int(result)


def sdiv(computation: ComputationAPI) -> None:
    """
    Signed Division
    """
    numerator, denominator = map(
        unsigned_to_signed,
        computation.stack_pop_ints(2),
    )

    pos_or_neg = -1 if numerator * denominator < 0 else 1

    if denominator == 0:
        result = 0
    else:
        result = pos_or_neg * (abs(numerator) // abs(denominator))

    computation.stack_push_int(signed_to_unsigned(result))


@curry
def exp(computation: ComputationAPI, gas_per_byte: int) -> None:
    """
    Exponentiation
    """
    base, exponent = computation.stack_pop_ints(2)

    bit_size = exponent.bit_length()
    byte_size = ceil8(bit_size) // 8

    if exponent == 0:
        result = 1
    elif base == 0:
        result = 0
    else:
        result = pow(base, exponent, constants.UINT_256_CEILING)

    computation.consume_gas(
        gas_per_byte * byte_size,
        reason="EXP: exponent bytes",
    )

    computation.stack_push_int(result)


def signextend(computation: ComputationAPI) -> None:
    """
    Signed Extend
    """
    bits, value = computation.stack_pop_ints(2)

    if bits <= 31:
        testbit = bits * 8 + 7
        sign_bit = 1 << testbit
        if value & sign_bit:
            result = value | (constants.UINT_256_CEILING - sign_bit)
        else:
            result = value & (sign_bit - 1)
    else:
        result = value

    computation.stack_push_int(result)


def shl(computation: ComputationAPI) -> None:
    """
    Bitwise left shift
    """
    shift_length, value = computation.stack_pop_ints(2)

    if shift_length >= 256:
        result = 0
    else:
        result = (value << shift_length) & constants.UINT_256_MAX

    computation.stack_push_int(result)


def shr(computation: ComputationAPI) -> None:
    """
    Bitwise right shift
    """
    shift_length, value = computation.stack_pop_ints(2)

    if shift_length >= 256:
        result = 0
    else:
        result = (value >> shift_length) & constants.UINT_256_MAX

    computation.stack_push_int(result)


def sar(computation: ComputationAPI) -> None:
    """
    Arithmetic bitwise right shift
    """
    shift_length, value = computation.stack_pop_ints(2)
    value = unsigned_to_signed(value)

    if shift_length >= 256:
        result = 0 if value >= 0 else constants.UINT_255_NEGATIVE_ONE
    else:
        result = (value >> shift_length) & constants.UINT_256_MAX

    computation.stack_push_int(result)



#
# EIP150 Mainnet Block
#
TANGERINE_WHISTLE_MAINNET_BLOCK = 2463000


#
# Gas Costs and Refunds
#
GAS_SELFDESTRUCT_EIP150 = 5000
GAS_CALL_EIP150 = 700
GAS_EXTCODE_EIP150 = 700
GAS_BALANCE_EIP150 = 400
GAS_SLOAD_EIP150 = 200



from typing import (
    Iterable,
    Optional,
    Tuple,
)

from eth_utils import (
    ValidationError,
    to_tuple,
)
from eth_utils.toolz import (
    curry,
)
import rlp

from eth.rlp.blocks import (
    BaseBlock,
)


@to_tuple
def diff_rlp_object(
    left: BaseBlock, right: BaseBlock
) -> Optional[Iterable[Tuple[str, str, str]]]:
    if left != right:
        rlp_type = type(left)

        for field_name, field_type in rlp_type._meta.fields:
            left_value = getattr(left, field_name)
            right_value = getattr(right, field_name)
            if isinstance(field_type, type) and issubclass(
                field_type, rlp.Serializable
            ):
                sub_diff = diff_rlp_object(left_value, right_value)
                for sub_field_name, sub_left_value, sub_right_value in sub_diff:
                    yield (
                        f"{field_name}.{sub_field_name}",
                        sub_left_value,
                        sub_right_value,
                    )
            elif isinstance(field_type, (rlp.sedes.List, rlp.sedes.CountableList)):
                if tuple(left_value) != tuple(right_value):
                    yield (
                        field_name,
                        left_value,
                        right_value,
                    )
            elif left_value != right_value:
                yield (
                    field_name,
                    left_value,
                    right_value,
                )
            else:
                continue


def _humanized_diff_elements(
    diff: Iterable[Tuple[str, str, str]], obj_a_name: str, obj_b_name: str
) -> Iterable[str]:
    longest_obj_name = max(len(obj_a_name), len(obj_b_name))

    for field_name, a_val, b_val in diff:
        if isinstance(a_val, int) and isinstance(b_val, int):
            element_diff = b_val - a_val
            if element_diff > 0:
                element_diff_display = f" (+{element_diff})"
            else:
                element_diff_display = f" ({element_diff})"
        else:
            element_diff_display = ""

        yield (
            f"{field_name}:\n"
            f"    ({obj_a_name.ljust(longest_obj_name, ' ')}) : {a_val}\n"
            f"    ({obj_b_name.ljust(longest_obj_name, ' ')}) : {b_val}{element_diff_display}"  # noqa: E501
        )


@curry
def validate_rlp_equal(
    obj_a: BaseBlock, obj_b: BaseBlock, obj_a_name: str = None, obj_b_name: str = None
) -> None:
    if obj_a == obj_b:
        return

    if obj_a_name is None:
        obj_a_name = obj_a.__class__.__name__ + "_a"
    if obj_b_name is None:
        obj_b_name = obj_b.__class__.__name__ + "_b"

    diff = diff_rlp_object(obj_a, obj_b)
    if len(diff) == 0:
        raise TypeError(
            f"{obj_a_name} ({obj_a!r}) != "
            f"{obj_b_name} ({obj_b!r}) but got an empty diff"
        )

    err_fields = "\n - ".join(_humanized_diff_elements(diff, obj_a_name, obj_b_name))
    error_message = (
        f"Mismatch between {obj_a_name} and {obj_b_name} "
        f"on {len(diff)} fields:\n - {err_fields}"
    )
    raise ValidationError(error_message)


validate_imported_block_unchanged = validate_rlp_equal(
    obj_a_name="locally executed block",
    obj_b_name="proposed block",
)



from eth_typing import (
    Hash32,
)

from eth.abc import (
    BlockHeaderAPI,
    ChainDatabaseAPI,
    StateAPI,
)
from eth.typing import (
    AccountState,
)


def get_parent_header(
    block_header: BlockHeaderAPI, db: ChainDatabaseAPI
) -> BlockHeaderAPI:
    """
    Returns the header for the parent block.
    """
    return db.get_block_header_by_hash(block_header.parent_hash)


def get_block_header_by_hash(
    block_hash: Hash32, db: ChainDatabaseAPI
) -> BlockHeaderAPI:
    """
    Returns the header for the parent block.
    """
    return db.get_block_header_by_hash(block_hash)


def apply_state_dict(state: StateAPI, state_dict: AccountState) -> None:
    for account, account_data in state_dict.items():
        state.set_balance(account, account_data["balance"])
        state.set_nonce(account, account_data["nonce"])
        state.set_code(account, account_data["code"])

        for slot, value in account_data["storage"].items():
            state.set_storage(account, slot, value)



from typing import Type

from eth_utils.exceptions import ValidationError

from eth.abc import (
    BlockHeaderAPI,
)
from eth.rlp.blocks import BaseBlock
from eth.validation import validate_gas_limit
from eth.vm.forks.berlin import BerlinVM
from eth.vm.state import BaseState

from .blocks import LondonBlock
from .constants import (
    EIP3529_MAX_REFUND_QUOTIENT,
    ELASTICITY_MULTIPLIER,
)
from .headers import (
    calculate_expected_base_fee_per_gas,
    compute_london_difficulty,
    configure_london_header,
    create_london_header_from_parent,
)
from .state import LondonState


class LondonVM(BerlinVM):
    # fork name
    fork = "london"

    # classes
    block_class: Type[BaseBlock] = LondonBlock
    _state_class: Type[BaseState] = LondonState

    # Methods
    create_header_from_parent = staticmethod(  # type: ignore
        create_london_header_from_parent(compute_london_difficulty)
    )
    compute_difficulty = staticmethod(compute_london_difficulty)  # type: ignore
    configure_header = configure_london_header

    @classmethod
    def validate_gas(
        cls, header: BlockHeaderAPI, parent_header: BlockHeaderAPI
    ) -> None:
        if hasattr(parent_header, "base_fee_per_gas"):
            # Follow normal gas limit rules if the previous block had a base fee
            parent_gas_limit = parent_header.gas_limit
        else:
            # On the fork block, double the gas limit.
            # That way, the gas target (which is half the London limit) equals the
            # previous (pre-London) gas limit.
            parent_gas_limit = parent_header.gas_limit * ELASTICITY_MULTIPLIER

        validate_gas_limit(header.gas_limit, parent_gas_limit)

        expected_base_fee_per_gas = calculate_expected_base_fee_per_gas(parent_header)
        if expected_base_fee_per_gas != header.base_fee_per_gas:
            raise ValidationError(
                f"Header has invalid base fee per gas (has {header.base_fee_per_gas}"
                f", expected {expected_base_fee_per_gas})"
            )

    @classmethod
    def calculate_net_gas_refund(cls, consumed_gas: int, gross_refund: int) -> int:
        max_refund = consumed_gas // EIP3529_MAX_REFUND_QUOTIENT
        return min(max_refund, gross_refund)



from typing import (
    Iterable,
    Tuple,
    cast,
)

from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Hash32,
)
import rlp

from eth.rlp.logs import (
    Log,
)


def hash_log_entries(
    log_entries: Iterable[Tuple[bytes, Tuple[int, ...], bytes]]
) -> Hash32:
    """
    Helper function for computing the RLP hash of the logs from transaction
    execution.
    """
    logs = [Log(*entry) for entry in log_entries]
    encoded_logs = rlp.encode(logs)
    logs_hash = keccak(encoded_logs)
    return cast(Hash32, logs_hash)



import functools
from typing import (
    Dict,
    Sequence,
    Tuple,
    Union,
)

from eth_typing import (
    Hash32,
)
import rlp
from trie import (
    HexaryTrie,
)

from eth.abc import (
    ReceiptAPI,
    SignedTransactionAPI,
    WithdrawalAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
)

BlockRootData = Union[
    Sequence[ReceiptAPI], Sequence[SignedTransactionAPI], Sequence[WithdrawalAPI]
]
TrieRootAndData = Tuple[Hash32, Dict[Hash32, bytes]]


def make_trie_root_and_nodes(items: BlockRootData) -> TrieRootAndData:
    return _make_trie_root_and_nodes(tuple(item.encode() for item in items))


# This cache is expected to be useful when importing blocks as we call this once when'
# importing and again when validating the imported block. But it should also help for
# post-Byzantium blocks as it's common for them to have duplicate receipt_roots.
# Given that, it probably makes sense to use a relatively small cache size here.
@functools.lru_cache(128)
def _make_trie_root_and_nodes(items: Tuple[bytes, ...]) -> TrieRootAndData:
    kv_store: Dict[Hash32, bytes] = {}
    trie = HexaryTrie(kv_store, BLANK_ROOT_HASH)
    with trie.squash_changes() as memory_trie:
        for index, item in enumerate(items):
            index_key = rlp.encode(index, sedes=rlp.sedes.big_endian_int)
            memory_trie[index_key] = item
    return trie.root_hash, kv_store



from .sha256 import sha256  # noqa: F401
from .identity import identity  # noqa: F401
from .ecrecover import ecrecover  # noqa: F401
from .ripemd160 import ripemd160  # noqa: F401
from .modexp import modexp  # noqa: F401
from .ecadd import ecadd  # noqa: F401
from .ecmul import ecmul  # noqa: F401
from .ecpairing import ecpairing  # noqa: F401
from .blake2 import blake2b_fcompress  # noqa: F401



from contextlib import (
    contextmanager,
)
import logging
from pathlib import (
    Path,
)
from typing import (
    TYPE_CHECKING,
    Iterator,
)

from eth_utils import (
    ValidationError,
)

from eth._warnings import (
    catch_and_ignore_import_warning,
)
from eth.abc import (
    AtomicWriteBatchAPI,
    DatabaseAPI,
)
from eth.db.diff import (
    DBDiffTracker,
    DiffMissingError,
)

from .base import (
    BaseAtomicDB,
    BaseDB,
)

if TYPE_CHECKING:
    with catch_and_ignore_import_warning():
        import plyvel  # noqa: F401


class LevelDB(BaseAtomicDB):
    logger = logging.getLogger("eth.db.backends.LevelDB")

    # Creates db as a class variable to avoid level db lock error
    def __init__(self, db_path: Path = None, max_open_files: int = None) -> None:
        if not db_path:
            raise TypeError("Please specifiy a valid path for your database.")
        try:
            with catch_and_ignore_import_warning():
                import plyvel  # noqa: F811
        except ImportError:
            raise ImportError(
                "LevelDB requires the plyvel library which is not available for import."
            )
        self.db_path = db_path
        self.db = plyvel.DB(
            str(db_path),
            create_if_missing=True,
            error_if_exists=False,
            max_open_files=max_open_files,
        )

    def __getitem__(self, key: bytes) -> bytes:
        v = self.db.get(key)
        if v is None:
            raise KeyError(key)
        return v

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.db.put(key, value)

    def _exists(self, key: bytes) -> bool:
        return self.db.get(key) is not None

    def __delitem__(self, key: bytes) -> None:
        if self.db.get(key) is None:
            raise KeyError(key)
        self.db.delete(key)

    @contextmanager
    def atomic_batch(self) -> Iterator[AtomicWriteBatchAPI]:
        with self.db.write_batch(transaction=True) as atomic_batch:
            readable_batch = LevelDBWriteBatch(self, atomic_batch)
            try:
                yield readable_batch
            finally:
                readable_batch.decommission()


class LevelDBWriteBatch(BaseDB, AtomicWriteBatchAPI):
    """
    A native leveldb write batch does not permit reads on the in-progress data.
    This class fills that gap, by tracking the in-progress diff, and adding
    a read interface.
    """

    logger = logging.getLogger("eth.db.backends.LevelDBWriteBatch")

    def __init__(
        self, original_read_db: DatabaseAPI, write_batch: "plyvel.WriteBatch"
    ) -> None:
        self._original_read_db = original_read_db
        self._write_batch = write_batch
        # keep track of the temporary changes made
        self._track_diff = DBDiffTracker()

    def __getitem__(self, key: bytes) -> bytes:
        if self._track_diff is None:
            raise ValidationError("Cannot get data from a write batch, out of context")

        try:
            changed_value = self._track_diff[key]
        except DiffMissingError as missing:
            if missing.is_deleted:
                raise KeyError(key)
            else:
                return self._original_read_db[key]
        else:
            return changed_value

    def __setitem__(self, key: bytes, value: bytes) -> None:
        if self._track_diff is None:
            raise ValidationError("Cannot set data from a write batch, out of context")

        self._write_batch.put(key, value)
        self._track_diff[key] = value

    def _exists(self, key: bytes) -> bool:
        if self._track_diff is None:
            raise ValidationError(
                "Cannot test data existance from a write batch, out of context"
            )

        try:
            self._track_diff[key]
        except DiffMissingError as missing:
            if missing.is_deleted:
                return False
            else:
                return key in self._original_read_db
        else:
            return True

    def __delitem__(self, key: bytes) -> None:
        if self._track_diff is None:
            raise ValidationError(
                "Cannot delete data from a write batch, out of context"
            )

        self._write_batch.delete(key)
        del self._track_diff[key]

    def decommission(self) -> None:
        """
        Prevent any further actions to be taken on this write batch,
        called after leaving context
        """
        self._track_diff = None



from eth_utils import (
    ValidationError,
    replace_exceptions,
)

from eth._utils.rlp import (
    validate_rlp_equal,
)

assert_imported_block_unchanged = replace_exceptions(
    {
        ValidationError: AssertionError,
    }
)(validate_rlp_equal(obj_a_name="provided block", obj_b_name="executed block"))


assert_headers_eq = replace_exceptions(
    {
        ValidationError: AssertionError,
    }
)(validate_rlp_equal(obj_a_name="expected", obj_b_name="actual"))



from eth.vm.forks.istanbul.state import (
    IstanbulState,
)

from .computation import (
    MuirGlacierComputation,
)


class MuirGlacierState(IstanbulState):
    computation_class = MuirGlacierComputation



from eth.vm.forks.byzantium.headers import (
    compute_difficulty,
    configure_header,
    create_header_from_parent,
)

compute_petersburg_difficulty = compute_difficulty(5000000)

create_petersburg_header_from_parent = create_header_from_parent(
    compute_petersburg_difficulty
)
configure_petersburg_header = configure_header(
    difficulty_fn=compute_petersburg_difficulty
)



import functools
import time
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Tuple,
    Type,
    Union,
    cast,
)

from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
)
from eth_utils import (
    ValidationError,
    to_dict,
    to_tuple,
)
from eth_utils.toolz import (
    curry,
    merge,
    pipe,
)

from eth import (
    constants,
)
from eth.abc import (
    AtomicDatabaseAPI,
    BlockAPI,
    ChainAPI,
    MiningChainAPI,
    VirtualMachineAPI,
)
from eth.consensus.applier import (
    ConsensusApplier,
)
from eth.consensus.noproof import (
    NoProofConsensus,
)
from eth.db.atomic import (
    AtomicDB,
)
from eth.db.backends.memory import (
    MemoryDB,
)
from eth.rlp.headers import (
    HeaderParams,
)
from eth.tools._utils.mappings import (
    deep_merge,
)
from eth.tools._utils.normalization import (
    normalize_state,
)
from eth.tools.mining import (
    POWMiningMixin,
)
from eth.typing import (
    AccountState,
    GeneralState,
    VMConfiguration,
    VMFork,
)
from eth.validation import (
    validate_vm_configuration,
)
from eth.vm.forks import (
    ArrowGlacierVM,
    BerlinVM,
    ByzantiumVM,
    ConstantinopleVM,
    FrontierVM,
    GrayGlacierVM,
    HomesteadVM,
    IstanbulVM,
    LondonVM,
    MuirGlacierVM,
    ParisVM,
    PetersburgVM,
    ShanghaiVM,
    SpuriousDragonVM,
    TangerineWhistleVM,
)


def build(obj: Any, *applicators: Callable[..., Any]) -> Any:
    """
    Run the provided object through the series of applicator functions.

    If ``obj`` is an instances of :class:`~eth.chains.base.BaseChain` the
    applicators will be run on a copy of the chain and thus will not mutate the
    provided chain instance.
    """
    if isinstance(obj, ChainAPI):
        return pipe(obj, copy(), *applicators)
    else:
        return pipe(obj, *applicators)


#
# Constructors (creation of chain classes)
#
@curry
def name(class_name: str, chain_class: Type[ChainAPI]) -> Type[ChainAPI]:
    """
    Assign the given name to the chain class.
    """
    return chain_class.configure(__name__=class_name)


@curry
def chain_id(chain_id: int, chain_class: Type[ChainAPI]) -> Type[ChainAPI]:
    """
    Set the ``chain_id`` for the chain class.
    """
    return chain_class.configure(chain_id=chain_id)


@curry
def fork_at(
    vm_class: Type[VirtualMachineAPI],
    at_block: Union[int, BlockNumber],
    chain_class: Type[ChainAPI],
) -> Type[ChainAPI]:
    """
    Adds the ``vm_class`` to the chain's ``vm_configuration``.

    .. code-block:: python

        from eth.chains.base import MiningChain
        from eth.tools.builder.chain import build, fork_at

        FrontierOnlyChain = build(MiningChain, fork_at(FrontierVM, 0))

        # these two classes are functionally equivalent.
        class FrontierOnlyChain(MiningChain):
            vm_configuration = (
                (0, FrontierVM),
            )

    .. note:: This function is curriable.

    The following pre-curried versions of this function are available as well,
    one for each mainnet fork.

    * :func:`~eth.tools.builder.chain.frontier_at`
    * :func:`~eth.tools.builder.chain.homestead_at`
    * :func:`~eth.tools.builder.chain.tangerine_whistle_at`
    * :func:`~eth.tools.builder.chain.spurious_dragon_at`
    * :func:`~eth.tools.builder.chain.byzantium_at`
    * :func:`~eth.tools.builder.chain.constantinople_at`
    * :func:`~eth.tools.builder.chain.petersburg_at`
    * :func:`~eth.tools.builder.chain.istanbul_at`
    * :func:`~eth.tools.builder.chain.muir_glacier_at`
    * :func:`~eth.tools.builder.chain.berlin_at`
    * :func:`~eth.tools.builder.chain.london_at`
    * :func:`~eth.tools.builder.chain.arrow_glacier_at`
    * :func:`~eth.tools.builder.chain.gray_glacier_at`
    * :func:`~eth.tools.builder.chain.latest_mainnet_at` - whatever latest mainnet VM is
    """
    if chain_class.vm_configuration is not None:
        base_configuration = chain_class.vm_configuration
    else:
        base_configuration = ()

    vm_configuration = base_configuration + ((BlockNumber(at_block), vm_class),)
    validate_vm_configuration(vm_configuration)
    return chain_class.configure(vm_configuration=vm_configuration)


def _is_homestead(vm_class: Type[VirtualMachineAPI]) -> bool:
    if not issubclass(vm_class, HomesteadVM):
        # It isn't a subclass of the HomesteadVM
        return False
    elif issubclass(vm_class, TangerineWhistleVM):
        # It is a subclass of one of the subsequent forks
        return False
    else:
        return True


@to_tuple
def _set_vm_dao_support_false(vm_configuration: VMConfiguration) -> Iterable[VMFork]:
    for fork_block, vm_class in vm_configuration:
        if _is_homestead(vm_class):
            yield fork_block, vm_class.configure(support_dao_fork=False)
        else:
            yield fork_block, vm_class


@curry
def disable_dao_fork(chain_class: Type[ChainAPI]) -> Type[ChainAPI]:
    """
    Set the ``support_dao_fork`` flag to ``False`` on the
    :class:`~eth.vm.forks.homestead.HomesteadVM`.  Requires that presence of
    the :class:`~eth.vm.forks.homestead.HomesteadVM`  in the
    ``vm_configuration``
    """
    homstead_vms_found = any(
        _is_homestead(vm_class) for _, vm_class in chain_class.vm_configuration
    )
    if not homstead_vms_found:
        raise ValidationError("No HomesteadVM found in vm_configuration.")

    vm_configuration = _set_vm_dao_support_false(chain_class.vm_configuration)
    return chain_class.configure(vm_configuration=vm_configuration)


@to_tuple
def _set_vm_dao_fork_block_number(
    dao_fork_block_number: BlockNumber, vm_configuration: VMConfiguration
) -> Iterable[VMFork]:
    for fork_block, vm_class in vm_configuration:
        if _is_homestead(vm_class):
            yield fork_block, vm_class.configure(
                support_dao_fork=True,
                _dao_fork_block_number=dao_fork_block_number,
            )
        else:
            yield fork_block, vm_class


@curry
def dao_fork_at(
    dao_fork_block_number: BlockNumber, chain_class: Type[ChainAPI]
) -> Type[ChainAPI]:
    """
    Set the block number on which the DAO fork will happen.  Requires that a
    version of the :class:`~eth.vm.forks.homestead.HomesteadVM` is present in
    the chain's ``vm_configuration``

    """
    homstead_vms_found = any(
        _is_homestead(vm_class) for _, vm_class in chain_class.vm_configuration
    )
    if not homstead_vms_found:
        raise ValidationError("No HomesteadVM found in vm_configuration.")

    vm_configuration = _set_vm_dao_fork_block_number(
        dao_fork_block_number,
        chain_class.vm_configuration,
    )
    return chain_class.configure(vm_configuration=vm_configuration)


frontier_at = fork_at(FrontierVM)
homestead_at = fork_at(HomesteadVM)
tangerine_whistle_at = fork_at(TangerineWhistleVM)
spurious_dragon_at = fork_at(SpuriousDragonVM)
byzantium_at = fork_at(ByzantiumVM)
constantinople_at = fork_at(ConstantinopleVM)
petersburg_at = fork_at(PetersburgVM)
istanbul_at = fork_at(IstanbulVM)
muir_glacier_at = fork_at(MuirGlacierVM)
berlin_at = fork_at(BerlinVM)
london_at = fork_at(LondonVM)
arrow_glacier_at = fork_at(ArrowGlacierVM)
gray_glacier_at = fork_at(GrayGlacierVM)
paris_at = fork_at(ParisVM)
shanghai_at = fork_at(ShanghaiVM)

latest_mainnet_at = shanghai_at

GENESIS_DEFAULTS = cast(
    Tuple[Tuple[str, Union[BlockNumber, int, None, bytes, Address, Hash32]], ...],
    # values that will automatically be default are commented out
    (
        ("difficulty", 1),
        ("extra_data", constants.GENESIS_EXTRA_DATA),
        ("gas_limit", constants.GENESIS_GAS_LIMIT),
        # ('gas_used', 0),
        # ('bloom', 0),
        ("mix_hash", constants.ZERO_HASH32),
        ("nonce", constants.GENESIS_NONCE),
        # ('block_number', constants.GENESIS_BLOCK_NUMBER),
        # ('parent_hash', constants.GENESIS_PARENT_HASH),
        ("receipt_root", constants.BLANK_ROOT_HASH),
        # ('uncles_hash', constants.EMPTY_UNCLE_HASH),
        ("state_root", constants.BLANK_ROOT_HASH),
        ("transaction_root", constants.BLANK_ROOT_HASH),
    ),
)


@to_dict
def _get_default_genesis_params(
    genesis_state: AccountState,
) -> Iterable[Tuple[str, Union[BlockNumber, int, None, bytes, Address, Hash32]]]:
    for key, value in GENESIS_DEFAULTS:
        if key == "state_root" and genesis_state:
            # leave out the `state_root` if a genesis state was specified
            pass
        else:
            yield key, value
    yield "timestamp", int(time.time())  # populate the timestamp value at runtime


@to_tuple
def _mix_in_pow_mining(vm_configuration: VMConfiguration) -> Iterable[VMFork]:
    for fork_block, vm_class in vm_configuration:
        vm_class_with_pow_mining = type(
            vm_class.__name__, (POWMiningMixin, vm_class), {}
        )
        yield fork_block, vm_class_with_pow_mining


@curry
def enable_pow_mining(chain_class: Type[ChainAPI]) -> Type[ChainAPI]:
    """
    Inject on demand generation of the proof of work mining seal on newly
    mined blocks into each of the chain's vms.
    """
    if not chain_class.vm_configuration:
        raise ValidationError("Chain class has no vm_configuration")

    vm_configuration = _mix_in_pow_mining(chain_class.vm_configuration)
    return chain_class.configure(vm_configuration=vm_configuration)


@curry
def disable_pow_check(chain_class: Type[ChainAPI]) -> Type[ChainAPI]:
    """
    Disable the proof of work validation check for each of the chain's vms.
    This allows for block mining without generation of the proof of work seal.

    .. note::

        blocks mined this way will not be importable on any chain that does not
        have proof of work disabled.
    """
    original_vms = chain_class.vm_configuration
    no_pow_vms = ConsensusApplier(NoProofConsensus).amend_vm_configuration(original_vms)

    return chain_class.configure(vm_configuration=no_pow_vms)


#
# Initializers (initialization of chain state and chain class instantiation)
#
def _fill_and_normalize_state(simple_state: GeneralState) -> AccountState:
    base_state = normalize_state(simple_state)
    defaults = {
        address: {
            "balance": 0,
            "nonce": 0,
            "code": b"",
            "storage": {},
        }
        for address in base_state.keys()
    }
    state = deep_merge(defaults, base_state)
    return state


@curry
def genesis(
    chain_class: ChainAPI,
    db: AtomicDatabaseAPI = None,
    params: Dict[str, HeaderParams] = None,
    state: GeneralState = None,
) -> ChainAPI:
    """
    Initialize the given chain class with the given genesis header parameters
    and chain state.
    """
    if state is None:
        genesis_state: AccountState = {}
    else:
        genesis_state = _fill_and_normalize_state(state)

    genesis_params_defaults = _get_default_genesis_params(genesis_state)

    if params is None:
        genesis_params = genesis_params_defaults
    else:
        genesis_params = merge(genesis_params_defaults, params)

    if db is None:
        base_db: AtomicDatabaseAPI = AtomicDB()
    else:
        base_db = db

    return chain_class.from_genesis(base_db, genesis_params, genesis_state)


#
# Builders (build actual blockchain)
#
@curry
def mine_block(chain: MiningChainAPI, **kwargs: Any) -> MiningChainAPI:
    """
    Mine a new block on the chain.  Header parameters for the new block can be
    overridden using keyword arguments.

    """

    if not isinstance(chain, MiningChainAPI):
        raise ValidationError("`mine_block` may only be used on MiningChain instances")

    transactions = kwargs.pop("transactions", ())

    chain.mine_all(transactions, **kwargs)
    return chain


@curry
def mine_blocks(num_blocks: int, chain: MiningChainAPI) -> MiningChainAPI:
    """
    Variadic argument version of :func:`~eth.tools.builder.chain.mine_block`
    """
    if not isinstance(chain, MiningChainAPI):
        raise ValidationError("`mine_block` may only be used on MiningChain instances")
    for _ in range(num_blocks):
        chain.mine_block()
    return chain


@curry
def import_block(block: BlockAPI, chain: ChainAPI) -> ChainAPI:
    """
    Import the provided ``block`` into the chain.
    """
    chain.import_block(block)
    return chain


def import_blocks(*blocks: BlockAPI) -> Callable[[ChainAPI], ChainAPI]:
    """
    Variadic argument version of :func:`~eth.tools.builder.chain.import_block`
    """

    @functools.wraps(import_blocks)
    def _import_blocks(chain: ChainAPI) -> ChainAPI:
        for block in blocks:
            chain.import_block(block)
        return chain

    return _import_blocks


@curry
def copy(chain: MiningChainAPI) -> MiningChainAPI:
    """
    Make a copy of the chain at the given state.  Actions performed on the
    resulting chain will not affect the original chain.
    """
    if not isinstance(chain, MiningChainAPI):
        raise ValidationError("`at_block_number` may only be used with 'MiningChain")
    base_db = chain.chaindb.db
    if not isinstance(base_db, AtomicDB):
        raise ValidationError(f"Unsupported database type: {type(base_db)}")

    if isinstance(base_db.wrapped_db, MemoryDB):
        db = AtomicDB(MemoryDB(base_db.wrapped_db.kv_store.copy()))
    else:
        raise ValidationError(
            f"Unsupported wrapped database: {type(base_db.wrapped_db)}"
        )

    chain_copy = type(chain)(db, chain.header)
    return chain_copy


def chain_split(
    *splits: Iterable[Callable[..., Any]]
) -> Callable[[ChainAPI], Iterable[ChainAPI]]:
    """
    Construct and execute multiple concurrent forks of the chain.

    Any number of forks may be executed.  For each fork, provide an iterable of
    commands.

    Returns the resulting chain objects for each fork.


    .. code-block:: python

        chain_a, chain_b = build(
            mining_chain,
            chain_split(
                (mine_block(extra_data=b'chain-a'), mine_block()),
                (mine_block(extra_data=b'chain-b'), mine_block(), mine_block()),
            ),
        )
    """
    if not splits:
        raise ValidationError(
            "Cannot use `chain_split` without providing at least one split"
        )

    @functools.wraps(chain_split)
    @to_tuple
    def _chain_split(chain: ChainAPI) -> Iterable[ChainAPI]:
        for split_fns in splits:
            result = build(
                chain,
                *split_fns,
            )
            yield result

    return _chain_split


@curry
def at_block_number(
    block_number: Union[int, BlockNumber], chain: MiningChainAPI
) -> MiningChainAPI:
    """
    Rewind the chain back to the given block number.  Calls to things like
    ``get_canonical_head`` will still return the canonical head of the chain,
    however, you can use ``mine_block`` to mine fork chains.
    """
    if not isinstance(chain, MiningChainAPI):
        raise ValidationError("`at_block_number` may only be used with 'MiningChain")
    at_block = chain.get_canonical_block_by_number(BlockNumber(block_number))

    db = chain.chaindb.db
    chain_at_block = type(chain)(db, chain.create_header_from_parent(at_block.header))
    return chain_at_block



from eth_utils.toolz import (
    partial,
)

from eth.vm.forks.berlin import (
    constants as berlin_constants,
)
from eth.vm.forks.berlin.logic import (
    GAS_SCHEDULE_EIP2929,
    sstore_eip2929_generic,
)

SSTORE_CLEARS_SCHEDULE_EIP_3529 = (
    GAS_SCHEDULE_EIP2929.sstore_reset_gas
    + berlin_constants.ACCESS_LIST_STORAGE_KEY_COST_EIP_2930
)


GAS_SCHEDULE_EIP3529 = GAS_SCHEDULE_EIP2929._replace(
    sstore_clears_schedule=SSTORE_CLEARS_SCHEDULE_EIP_3529
)

sstore_eip3529 = partial(sstore_eip2929_generic, GAS_SCHEDULE_EIP3529)



from rlp.sedes import (
    CountableList,
)

from eth.rlp.headers import (
    BlockHeader,
)
from eth.vm.forks.byzantium.blocks import (
    ByzantiumBlock,
)

from .transactions import (
    ConstantinopleTransaction,
)


class ConstantinopleBlock(ByzantiumBlock):
    transaction_builder = ConstantinopleTransaction
    fields = [
        ("header", BlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(BlockHeader)),
    ]



from eth import (
    constants,
)
from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    ComputationAPI,
)


def identity(computation: ComputationAPI) -> ComputationAPI:
    word_count = ceil32(len(computation.msg.data)) // 32
    gas_fee = constants.GAS_IDENTITY + word_count * constants.GAS_IDENTITYWORD

    computation.consume_gas(gas_fee, reason="Identity Precompile")

    computation.output = computation.msg.data_as_bytes
    return computation



from collections import (
    OrderedDict,
)
from typing import (
    Iterable,
    Tuple,
)

from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    big_endian_to_int,
    encode_hex,
)
from pyethash import (
    EPOCH_LENGTH,
    hashimoto_light,
    mkcache_bytes,
)

from eth.abc import (
    AtomicDatabaseAPI,
    BlockHeaderAPI,
    ConsensusAPI,
)
from eth.validation import (
    validate_length,
    validate_lte,
)

# Type annotation here is to ensure we don't accidentally use strings instead of bytes.
cache_by_epoch: "OrderedDict[int, bytearray]" = OrderedDict()
CACHE_MAX_ITEMS = 10


def get_cache(block_number: int) -> bytes:
    epoch_index = block_number // EPOCH_LENGTH

    # doing explicit caching, because functools.lru_cache is 70% slower in the tests

    # Get the cache if already generated, marking it as recently used
    if epoch_index in cache_by_epoch:
        c = cache_by_epoch.pop(epoch_index)  # pop and append at end
        cache_by_epoch[epoch_index] = c
        return c

    # Generate the cache if it was not already in memory
    # Simulate requesting mkcache by block number: multiply index by epoch length
    c = mkcache_bytes(epoch_index * EPOCH_LENGTH)
    cache_by_epoch[epoch_index] = c

    # Limit memory usage for cache
    if len(cache_by_epoch) > CACHE_MAX_ITEMS:
        cache_by_epoch.popitem(last=False)  # remove last recently accessed

    return c


def check_pow(
    block_number: int,
    mining_hash: Hash32,
    mix_hash: Hash32,
    nonce: bytes,
    difficulty: int,
) -> None:
    validate_length(mix_hash, 32, title="Mix Hash")
    validate_length(mining_hash, 32, title="Mining Hash")
    validate_length(nonce, 8, title="POW Nonce")
    cache = get_cache(block_number)
    mining_output = hashimoto_light(
        block_number, cache, mining_hash, big_endian_to_int(nonce)
    )
    if mining_output[b"mix digest"] != mix_hash:
        raise ValidationError(
            f"mix hash mismatch; expected: {encode_hex(mining_output[b'mix digest'])} "
            f"!= actual: {encode_hex(mix_hash)}. "
            f"Mix hash calculated from block #{block_number}, "
            f"mine hash {encode_hex(mining_hash)}, nonce {encode_hex(nonce)}"
            f", difficulty {difficulty}, cache hash {encode_hex(keccak(cache))}"
        )
    result = big_endian_to_int(mining_output[b"result"])
    validate_lte(result, 2**256 // difficulty, title="POW Difficulty")


MAX_TEST_MINE_ATTEMPTS = 1000


def mine_pow_nonce(
    block_number: int, mining_hash: Hash32, difficulty: int
) -> Tuple[bytes, bytes]:
    cache = get_cache(block_number)
    for nonce in range(MAX_TEST_MINE_ATTEMPTS):
        mining_output = hashimoto_light(block_number, cache, mining_hash, nonce)
        result = big_endian_to_int(mining_output[b"result"])
        result_cap = 2**256 // difficulty
        if result <= result_cap:
            return nonce.to_bytes(8, "big"), mining_output[b"mix digest"]

    raise Exception("Too many attempts at POW mining, giving up")


class PowConsensus(ConsensusAPI):
    """
    Modify a set of VMs to validate blocks via Proof of Work (POW)
    """

    def __init__(self, base_db: AtomicDatabaseAPI) -> None:
        pass

    def validate_seal(self, header: BlockHeaderAPI) -> None:
        """
        Validate the seal on the given header by checking the proof of work.
        """
        check_pow(
            header.block_number,
            header.mining_hash,
            header.mix_hash,
            header.nonce,
            header.difficulty,
        )

    def validate_seal_extension(
        self, header: BlockHeaderAPI, parents: Iterable[BlockHeaderAPI]
    ) -> None:
        pass

    @classmethod
    def get_fee_recipient(cls, header: BlockHeaderAPI) -> Address:
        """
        Return the ``coinbase`` of the passed ``header`` as the receipient for any
        rewards for the block.
        """
        return header.coinbase



import os
from typing import Any, Type, cast

from eth_utils import import_string

from eth.abc import AtomicDatabaseAPI


DEFAULT_DB_BACKEND = "eth.db.atomic.AtomicDB"


def get_db_backend_class(import_path: str = None) -> Type[AtomicDatabaseAPI]:
    if import_path is None:
        import_path = os.environ.get(
            "CHAIN_DB_BACKEND_CLASS",
            DEFAULT_DB_BACKEND,
        )
    return cast(Type[AtomicDatabaseAPI], import_string(import_path))


def get_db_backend(import_path: str = None, **init_kwargs: Any) -> AtomicDatabaseAPI:
    backend_class = get_db_backend_class(import_path)
    # mypy doesn't understand the constructor  of AtomicDatabaseAPI
    return backend_class(**init_kwargs)



from eth.vm.forks.gray_glacier.computation import (
    GrayGlacierComputation,
)

from .opcodes import (
    PARIS_OPCODES,
)


class ParisComputation(GrayGlacierComputation):
    """
    A class for all execution *message* computations in the ``Paris`` hard fork
    Inherits from :class:`~eth.vm.forks.gray_glacier.GrayGlacierComputation`
    """

    opcodes = PARIS_OPCODES



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Tuple,
)

from eth_typing import (
    Address,
)

from eth import (
    constants,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth.abc import (
    ComputationAPI,
)
from eth.exceptions import (
    OutOfGas,
    WriteProtection,
)
from eth.vm.opcode import (
    Opcode,
)

CallParams = Tuple[int, int, Address, Address, Address, int, int, int, int, bool, bool]


class BaseCall(Opcode, ABC):
    @abstractmethod
    def compute_msg_extra_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> int:
        raise NotImplementedError("Must be implemented by subclasses")

    @abstractmethod
    def get_call_params(self, computation: ComputationAPI) -> CallParams:
        raise NotImplementedError("Must be implemented by subclasses")

    def compute_msg_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> Tuple[int, int]:
        extra_gas = self.compute_msg_extra_gas(computation, gas, to, value)
        total_fee = gas + extra_gas
        child_msg_gas = gas + (constants.GAS_CALLSTIPEND if value else 0)
        return child_msg_gas, total_fee

    def get_account_load_fee(
        self,
        computation: ComputationAPI,
        code_address: Address,
    ) -> int:
        """
        Return the gas cost for implicitly loading the account needed to access
        the bytecode.
        """
        return 0

    def __call__(self, computation: ComputationAPI) -> None:
        computation.consume_gas(
            self.gas_cost,
            reason=self.mnemonic,
        )

        (
            gas,
            value,
            to,
            sender,
            code_address,
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
            should_transfer_value,
            is_static,
        ) = self.get_call_params(computation)

        computation.extend_memory(memory_input_start_position, memory_input_size)
        computation.extend_memory(memory_output_start_position, memory_output_size)

        call_data = computation.memory_read(
            memory_input_start_position, memory_input_size
        )

        #
        # Message gas allocation and fees
        #
        if code_address:
            code_source = code_address
        else:
            code_source = to
        load_account_fee = self.get_account_load_fee(computation, code_source)
        if load_account_fee > 0:
            computation.consume_gas(
                load_account_fee,
                reason=f"{self.mnemonic} charges implicit account load for reading code",  # noqa: E501
            )
            if self.logger.show_debug2:
                self.logger.debug2(
                    "%s is charged %d for invoking code at account 0x%s",
                    self.mnemonic,
                    load_account_fee,
                    code_source.hex(),
                )

        # This must be computed *after* the load account fee is charged, so
        # that the 63/64ths rule is applied against the reduced remaining gas.
        child_msg_gas, child_msg_gas_fee = self.compute_msg_gas(
            computation, gas, to, value
        )
        computation.consume_gas(child_msg_gas_fee, reason=self.mnemonic)

        # Pre-call checks
        sender_balance = computation.state.get_balance(computation.msg.storage_address)

        insufficient_funds = should_transfer_value and sender_balance < value
        stack_too_deep = computation.msg.depth + 1 > constants.STACK_DEPTH_LIMIT

        if insufficient_funds or stack_too_deep:
            computation.return_data = b""
            if insufficient_funds:
                err_message = (
                    f"Insufficient Funds: have: {sender_balance} | need: {value}"
                )
            elif stack_too_deep:
                err_message = "Stack Limit Reached"
            else:
                raise Exception("Invariant: Unreachable code path")

            self.logger.debug2(
                "%s failure: %s",
                self.mnemonic,
                err_message,
            )
            computation.return_gas(child_msg_gas)
            computation.stack_push_int(0)
        else:
            if code_address:
                code = computation.state.get_code(code_address)
            else:
                code = computation.state.get_code(to)

            child_msg_kwargs = {
                "gas": child_msg_gas,
                "value": value,
                "to": to,
                "data": call_data,
                "code": code,
                "code_address": code_address,
                "should_transfer_value": should_transfer_value,
                "is_static": is_static,
            }
            if sender is not None:
                child_msg_kwargs["sender"] = sender

            # TODO: after upgrade to py3.6, use a TypedDict and try again
            child_msg = computation.prepare_child_message(**child_msg_kwargs)  # type: ignore  # noqa: E501

            child_computation = computation.apply_child_computation(child_msg)

            if child_computation.is_error:
                computation.stack_push_int(0)
            else:
                computation.stack_push_int(1)

            if not child_computation.should_erase_return_data:
                actual_output_size = min(
                    memory_output_size, len(child_computation.output)
                )
                computation.memory_write(
                    memory_output_start_position,
                    actual_output_size,
                    child_computation.output[:actual_output_size],
                )

            if child_computation.should_return_gas:
                computation.return_gas(child_computation.get_gas_remaining())


class Call(BaseCall):
    def compute_msg_extra_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> int:
        account_exists = computation.state.account_exists(to)

        transfer_gas_fee = constants.GAS_CALLVALUE if value else 0
        create_gas_fee = constants.GAS_NEWACCOUNT if not account_exists else 0
        return transfer_gas_fee + create_gas_fee

    def get_call_params(self, computation: ComputationAPI) -> CallParams:
        gas = computation.stack_pop1_int()
        to = force_bytes_to_address(computation.stack_pop1_bytes())

        (
            value,
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
        ) = computation.stack_pop_ints(5)

        return (
            gas,
            value,
            to,
            None,  # sender
            None,  # code_address
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
            True,  # should_transfer_value,
            computation.msg.is_static,
        )


class CallCode(BaseCall):
    def compute_msg_extra_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> int:
        return constants.GAS_CALLVALUE if value else 0

    def get_call_params(self, computation: ComputationAPI) -> CallParams:
        gas = computation.stack_pop1_int()
        code_address = force_bytes_to_address(computation.stack_pop1_bytes())

        (
            value,
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
        ) = computation.stack_pop_ints(5)

        to = computation.msg.storage_address
        sender = computation.msg.storage_address

        return (
            gas,
            value,
            to,
            sender,
            code_address,
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
            True,  # should_transfer_value,
            computation.msg.is_static,
        )


class DelegateCall(BaseCall):
    def compute_msg_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> Tuple[int, int]:
        return gas, gas

    def compute_msg_extra_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> int:
        return 0

    def get_call_params(self, computation: ComputationAPI) -> CallParams:
        gas = computation.stack_pop1_int()
        code_address = force_bytes_to_address(computation.stack_pop1_bytes())

        (
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
        ) = computation.stack_pop_ints(4)

        to = computation.msg.storage_address
        sender = computation.msg.sender
        value = computation.msg.value

        return (
            gas,
            value,
            to,
            sender,
            code_address,
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
            False,  # should_transfer_value,
            computation.msg.is_static,
        )


#
# EIP150
#
class CallEIP150(Call):
    def compute_msg_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> Tuple[int, int]:
        extra_gas = self.compute_msg_extra_gas(computation, gas, to, value)
        return compute_eip150_msg_gas(
            computation=computation,
            gas=gas,
            extra_gas=extra_gas,
            value=value,
            mnemonic=self.mnemonic,
            callstipend=constants.GAS_CALLSTIPEND,
        )


class CallCodeEIP150(CallCode):
    def compute_msg_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> Tuple[int, int]:
        extra_gas = self.compute_msg_extra_gas(computation, gas, to, value)
        return compute_eip150_msg_gas(
            computation=computation,
            gas=gas,
            extra_gas=extra_gas,
            value=value,
            mnemonic=self.mnemonic,
            callstipend=constants.GAS_CALLSTIPEND,
        )


class DelegateCallEIP150(DelegateCall):
    def compute_msg_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> Tuple[int, int]:
        extra_gas = self.compute_msg_extra_gas(computation, gas, to, value)
        callstipend = 0
        return compute_eip150_msg_gas(
            computation=computation,
            gas=gas,
            extra_gas=extra_gas,
            value=value,
            mnemonic=self.mnemonic,
            callstipend=callstipend,
        )


def max_child_gas_eip150(gas: int) -> int:
    return gas - (gas // 64)


def compute_eip150_msg_gas(
    *,
    computation: ComputationAPI,
    gas: int,
    extra_gas: int,
    value: int,
    mnemonic: str,
    callstipend: int,
) -> Tuple[int, int]:
    if computation.get_gas_remaining() < extra_gas:
        # It feels wrong to raise an OutOfGas exception outside of GasMeter,
        # but I don't see an easy way around it.
        raise OutOfGas(
            f"Out of gas: Needed {extra_gas}"
            f" - Remaining {computation.get_gas_remaining()}"
            f" - Reason: {mnemonic}"
        )
    gas = min(gas, max_child_gas_eip150(computation.get_gas_remaining() - extra_gas))
    total_fee = gas + extra_gas
    child_msg_gas = gas + (callstipend if value else 0)
    return child_msg_gas, total_fee


#
# EIP161
#
class CallEIP161(CallEIP150):
    def compute_msg_extra_gas(
        self, computation: ComputationAPI, gas: int, to: Address, value: int
    ) -> int:
        account_is_dead = not computation.state.account_exists(
            to
        ) or computation.state.account_is_empty(to)

        transfer_gas_fee = constants.GAS_CALLVALUE if value else 0
        create_gas_fee = constants.GAS_NEWACCOUNT if (account_is_dead and value) else 0
        return transfer_gas_fee + create_gas_fee


#
# Byzantium
#
class StaticCall(CallEIP161):
    def get_call_params(self, computation: ComputationAPI) -> CallParams:
        gas = computation.stack_pop1_int()
        to = force_bytes_to_address(computation.stack_pop1_bytes())

        (
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
        ) = computation.stack_pop_ints(4)

        return (
            gas,
            0,  # value
            to,
            None,  # sender
            None,  # code_address
            memory_input_start_position,
            memory_input_size,
            memory_output_start_position,
            memory_output_size,
            False,  # should_transfer_value,
            True,  # is_static
        )


class CallByzantium(CallEIP161):
    def get_call_params(self, computation: ComputationAPI) -> CallParams:
        call_params = super().get_call_params(computation)
        value = call_params[1]
        if computation.msg.is_static and value != 0:
            raise WriteProtection(
                "Cannot modify state while inside of a STATICCALL context"
            )
        return call_params



from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
)

from eth._utils.transactions import (
    create_transaction_signature,
)
from eth.vm.forks.byzantium.transactions import (
    ByzantiumTransaction,
    ByzantiumUnsignedTransaction,
)


class ConstantinopleTransaction(ByzantiumTransaction):
    @classmethod
    def create_unsigned_transaction(
        cls,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes
    ) -> "ConstantinopleUnsignedTransaction":
        return ConstantinopleUnsignedTransaction(nonce, gas_price, gas, to, value, data)


class ConstantinopleUnsignedTransaction(ByzantiumUnsignedTransaction):
    def as_signed_transaction(
        self, private_key: PrivateKey, chain_id: int = None
    ) -> ConstantinopleTransaction:
        v, r, s = create_transaction_signature(self, private_key, chain_id=chain_id)
        return ConstantinopleTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )



from functools import (
    partial,
)

from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
)
import rlp

from eth._utils.transactions import (
    calculate_intrinsic_gas,
    create_transaction_signature,
)
from eth.constants import (
    GAS_TXCREATE,
)
from eth.validation import (
    validate_lt_secpk1n2,
)
from eth.vm.forks.frontier.transactions import (
    FRONTIER_TX_GAS_SCHEDULE,
    FrontierTransaction,
    FrontierUnsignedTransaction,
)

HOMESTEAD_TX_GAS_SCHEDULE = FRONTIER_TX_GAS_SCHEDULE._replace(
    gas_txcreate=GAS_TXCREATE,
)


homestead_get_intrinsic_gas = partial(
    calculate_intrinsic_gas, HOMESTEAD_TX_GAS_SCHEDULE
)


class HomesteadTransaction(FrontierTransaction):
    def validate(self) -> None:
        super().validate()
        validate_lt_secpk1n2(self.s, title="Transaction.s")

    def get_intrinsic_gas(self) -> int:
        return homestead_get_intrinsic_gas(self)

    def get_message_for_signing(self) -> bytes:
        return rlp.encode(
            HomesteadUnsignedTransaction(
                nonce=self.nonce,
                gas_price=self.gas_price,
                gas=self.gas,
                to=self.to,
                value=self.value,
                data=self.data,
            )
        )

    @classmethod
    def create_unsigned_transaction(
        cls,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes
    ) -> "HomesteadUnsignedTransaction":
        return HomesteadUnsignedTransaction(nonce, gas_price, gas, to, value, data)


class HomesteadUnsignedTransaction(FrontierUnsignedTransaction):
    def as_signed_transaction(
        self,
        private_key: PrivateKey,
        chain_id: int = None,  # unused until SpuriousDragon
    ) -> HomesteadTransaction:
        v, r, s = create_transaction_signature(self, private_key)
        return HomesteadTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )

    def get_intrinsic_gas(self) -> int:
        return homestead_get_intrinsic_gas(self)



import itertools
from typing import (
    Generic,
    Iterable,
    Iterator,
    List,
    TypeVar,
)

TItem = TypeVar("TItem")


class CachedIterable(Generic[TItem], Iterable[TItem]):
    def __init__(self, iterable: Iterable[TItem]) -> None:
        self._cached_results: List[TItem] = []
        self._iterator = iter(iterable)

    def __iter__(self) -> Iterator[TItem]:
        return itertools.chain(self._cached_results, self._cache_and_yield())

    def _cache_and_yield(self) -> Iterator[TItem]:
        for item in self._iterator:
            self._cached_results.append(item)
            yield item
